from copy import deepcopy
from dataclasses import dataclass

import numpy as np


@dataclass
class BaseProperty:

    property: np.ndarray
    name: str
    framerate: int = None

    def __str__(self):
        return f"Floodlight {self.__class__.__name__} object encoding '{self.name}'"

    def __len__(self):
        return len(self.property)

    def __getitem__(self, key):
        return self.property[key]

    def __setitem__(self, key, value):
        self.property[key] = value

    @classmethod
    def _slice_new(cls, sliced_property, name, framerate):
        sliced_copy = cls(
            property=sliced_property,
            name=name,
            framerate=framerate,
        )

        return sliced_copy

    def slice(
        self, startframe: int = None, endframe: int = None, inplace: bool = False
    ):
        """Return copy of object with sliced property. Mimics numpy's array slicing.

        Parameters
        ----------
        startframe : int, optional
            Start of slice. Defaults to beginning of segment.
        endframe : int, optional
            End of slice (endframe is excluded). Defaults to end of segment.
        inplace: bool, optional
            If set to ``False`` (default), a new object is returned, otherwise the
            operation is performed in place on the called object.

        Returns
        -------
        property_sliced: Union[cls, None]
        """
        sliced_data = self.property[startframe:endframe].copy()
        sliced_object = None

        if inplace:
            self.property = sliced_data
        else:
            sliced_object = self._slice_new(
                sliced_property=sliced_data,
                name=deepcopy(self.name),
                framerate=deepcopy(self.framerate),
            )

        return sliced_object


@dataclass
class TeamProperty(BaseProperty):
    """Fragment of one continuous team property. Core class of floodlight.

    Attributes
    ----------
    property: np.ndarray
        A 1-dimensional array of properties of shape (T), where T is the number of
        total frames.
    name: str
        Name of the property (e.g. 'stretch_index').
    framerate: int, optional
        Temporal resolution of data in frames per second/Hertz.
    """


@dataclass
class PlayerProperty(BaseProperty):
    """Fragment of one continuous property per player. Core class of floodlight.

    Attributes
    ----------
    property: np.ndarray
        A 2-dimensional array of properties of shape (T, N), where T is the number of
        total frames and N is the number of players.
    name: str
        Name of the property (e.g. 'speed').
    framerate: int, optional
        Temporal resolution of data in frames per second/Hertz.
    """


@dataclass
class DyadicProperty(BaseProperty):
    """Fragment of one continuous property per player dyad. Core class of floodlight.

    Attributes
    ----------
    property: np.ndarray
        A 3-dimensional array of properties of shape (T, N_1, N_2), where T is the
        number of total frames and {N_1, N_2} are the number of players between dyads
        are formed. For example, the item at (1, 2, 3) encodes the relation from player
        with xID=2 to player with xID=3 at frame 1. Note that players could be in the
        same team (intra-team relations, in this case N_1 = N_2) or opposing teams
        (inter-team relations).
    name: str
        Name of the property (e.g. 'distance').
    framerate: int, optional
        Temporal resolution of data in frames per second/Hertz.
    """



import numpy as np
from scipy.constants import g

from floodlight.utils.types import Numeric
from floodlight.core.xy import XY
from floodlight.core.property import PlayerProperty
from floodlight.models.base import BaseModel, requires_fit
from floodlight.models.kinematics import VelocityModel, AccelerationModel


class MetabolicPowerModel(BaseModel):
    """Class for calculating Metabolic Power and derived metrics from spatiotemporal
    data.

    Upon calling the :func:`~MetbolicPowerModel.fit`-method, this model calculates the
    frame-wise Metabolic Power for each player. The following calculations can
    subsequently be queried by calling the corresponding methods:

        - Frame-wise Metabolic Power --> :func:`~MetabolicPowerModel.metabolic_power`
        - Cumulative Metabolic Power --> :func:`~MetabolicPowerModel.cumulative_\
metabolic_power`
        - Frame-wise Equivalent Distance --> :func:`~MetabolicPowerModel.equivalent_\
distance`
        - Cumulative Equivalent Distance --> :func:`~MetabolicPowerModel.cumulative_\
equivalent_distance`

    Notes
    -----
    Metabolic Power is defined as the energy expenditure over time necessary to move at
    a certain speed, and is calculated as the product of energy cost of transport per
    unit body mass and distance [:math:`\\frac{J}{kg \\cdot m}`] and velocity
    [:math:`\\frac{m}{s}`]. Metabolic Power and Energy cost of walking is calculated
    according to di Prampero & Osgnach [1]_. Energy cost of running is calculated with
    the updated formula of Minetti & Parvei [2]_.

    Examples
    --------
    >>> import numpy as np
    >>> from floodlight import XY
    >>> from floodlight.models.kinetics import MetabolicPowerModel

    >>> xy = XY(np.array(((0, 0), (0, 1), (1, 1), (2, 2))), framerate=20)

    >>> metabolic_power_model = MetabolicPowerModel()
    >>> metabolic_power_model.fit(xy)

    >>> metabolic_power_model.metabolic_power()
    PlayerProperty(property=array([[1164.59773017],
           [ 185.59792131],
           [9448.10007077],
           [8593.05199423]]), name='metabolic_power', framerate=20)
    >>> metabolic_power_model.cumulative_equivalent_distance()
    PlayerProperty(property=array([[ 323.49936949],
       [ 375.05434763],
       [2999.52658952],
       [5386.4854768 ]]), name='cumulative_equivalent_distance', framerate=20)

    References
    ----------
        .. [1] `di Prampero P.E., Osgnach C. (2018). Metabolic power in team sports -
            Part 1: An update. International Journal of Sports Medicine, 39(08),
            581-587.
            <https://www.thieme-connect.de/products/ejournals/abstract/10.1055/
            a-0592-7660>`_
        .. [2] `Minetti, A.E., Parvei, G. (2018). Update and extension of the
            ‘Equivalent Slope’ of speed changing level locomotion in humans: A
            computational model for shuttle running. Journal Experimental Biology,
            221:jeb.182303.
            <https://journals.biologists.com/jeb/article/221/15/jeb182303/19414/Update-
            and-extension-of-the-equivalent-slope-of>`_
    """

    # Coefficient of air resistance from di Prampero (2018).
    K = 0.0037

    # Coefficients of polynomial to calculate the walk-run-transition
    # velocity based on the equivalent slope from di Prampero (2018).
    RUNNING_TRANSITION_COEFF = np.array((-107.05, 113.13, -1.13, -15.84, -1.7, 2.27))

    # Cutoffs of equivalent slope for using the corresponding polynomial to calculate
    # energy cost of walking at a certain velocity from di Prampero (2018).
    ECW_ES_CUTOFFS = np.array([-0.3, -0.2, -0.1, 0, 0.1, 0.2, 0.3, 0.4])

    # Coefficients of polynomials to calculate energy cost of walking from di
    # Prampero (2018).
    ECW_POLY_COEFF = np.array(
        [
            [0.28, -1.66, 3.81, -3.96, 4.01],
            [0.03, -0.15, 0.98, -2.25, 3.14],
            [0.69, -3.21, 5.94, -5.07, 2.79],
            [1.25, -6.57, 13.14, -11.15, 5.35],
            [0.68, -4.17, 10.17, -10.31, 8.66],
            [3.80, -14.91, 22.94, -14.53, 11.24],
            [44.95, -122.88, 126.94, -57.46, 21.39],
            [94.62, -213.94, 184.43, -68.49, 25.04],
        ]
    )

    def __init__(self):
        super().__init__()
        self._metabolic_power_ = None

    @staticmethod
    def _calc_es(vel, acc):
        """Calculates equivalent slope based on the formula by di Prampero & Osgnach
        (2018)

        Parameters
        ----------
        vel: np.array
            velocity
        acc: np.array
            acceleration

        Returns
        -------
        es: np.array
            equivalent slope
        """
        es = (acc / g) + ((MetabolicPowerModel.K * np.square(vel)) / g)
        return es

    @staticmethod
    def _calc_em(es):
        """Calculates equivalent mass based on the formula by di Prampero & Osgnach
        (2018)

        Parameters
        ----------
        es: np.array
            equivalent slope

        Returns
        -------
        em: np.array
            equivalent mass
        """
        em = np.sqrt(np.square(es) + 1)
        return em

    @staticmethod
    def _calc_v_trans(es: np.ndarray) -> np.ndarray:
        """Calculate the walking to running transition velocity at a certain equivalent
        slope based on the formula of di Prampero (2018).

        Parameters
        ----------
        es: np.array
            equivalent slope

        Returns
        -------
        v_trans: np.array
            Array with the respective transition velocity

        """
        es_power = np.stack(
            (
                np.power(es, 5),
                np.power(es, 4),
                np.power(es, 3),
                np.power(es, 2),
                es,
                np.ones(es.shape),
            ),
            axis=-1,
        )

        v_trans = np.matmul(es_power, MetabolicPowerModel.RUNNING_TRANSITION_COEFF)

        return v_trans

    @staticmethod
    def _is_running(vel: np.ndarray, es: np.ndarray) -> np.ndarray:
        """
        Checks if athlete is walking or running based on the model of di Prampero
        (2018).

        Parameters
        ----------
        vel: np.array
            Velocity
        es: np.array
            Equivalent slope

        Returns
        -------
        is_running: np.ndarray
            Array containing boolean values indicating whether an athlete is running
            (True) or not (False).
        """
        # Calculate walk-run-transition velocity
        v_trans = MetabolicPowerModel._calc_v_trans(es)

        is_running = (vel >= v_trans) | (vel > 2.5)

        return is_running

    @staticmethod
    def _get_interpolation_weight_matrix(es: np.ndarray) -> np.ndarray:
        """Calculates interpolation weight matrix. This matrix is designed for a
        calculation of ECW in a single sweep by determining the interpolation weights
        of all 8 ECW_ES_CUTOFFS for given ES values.

        Parameters
        ----------
        es: np.array
            Equivalent slope

        Returns
        -------
        W: np.array
            Interpolation weight matrix of shape (T frames, N players,
            len(ECW_ES_CUTOFFS)=8) containing interpolation coefficients from range
            [0, 1].
        """
        # Number of frames
        T = es.shape[0]
        # Number of players
        N = es.shape[1]
        # Pre-allocated interpolation weight matrix with 3 dimensions (T frames, N
        # players, len(CUTOFFS)=8 polynomials)
        W = np.zeros((T, N, len(MetabolicPowerModel.ECW_ES_CUTOFFS)))

        # Index of each ES regarding its position in CUTOFFS.
        # E.g. es = 0.25 -> es will be sorted between CUTOFFS[5] and CUTOFFS[6],
        # idxs = 6
        idxs = MetabolicPowerModel.ECW_ES_CUTOFFS.searchsorted(es)

        # Mask for non-edge cases (es outside of CUTOFFS)
        mask = (idxs > 0) & (idxs < 8)

        # Initialize grids for appropriate indexing of W along axis=0 (time) and
        # axis=1 (player)
        grid_t, grid_n = np.mgrid[0:T, 0:N]

        # Fill W with the right interpolation weights for each time t (axis=0),
        # player n (axis=1) and polynomial (axis=2)
        W[grid_t[mask], grid_n[mask], idxs[mask] - 1] = (
            MetabolicPowerModel.ECW_ES_CUTOFFS[idxs[mask]] - es[mask]
        ) * 10
        W[grid_t[mask], grid_n[mask], idxs[mask]] = (
            es[mask] - MetabolicPowerModel.ECW_ES_CUTOFFS[idxs[mask] - 1]
        ) * 10

        # Fill edge cases (es not in range of CUTOFFS) with 1 because they are
        # calculated with the corresponding min/max CUTOFFS
        W[idxs == 0, 0] = 1
        W[idxs == 8, 7] = 1

        return W

    @staticmethod
    def _calc_ecw(es: np.ndarray, vel: np.ndarray, em: np.ndarray) -> np.ndarray:
        """Calculates energy cost of walking based on formula (13), (14) and table
        1 in di Prampero & Osgnach (2018).

        Parameters
        ----------
        es: np.array
            Equivalent slope
        vel: np.array
            Velocity
        em: np.array
            Equivalent mass

        Returns
        -------
        ECW: np.array
            Energy cost of walking

        """
        # Interpolation weight matrix
        W = MetabolicPowerModel._get_interpolation_weight_matrix(es)

        # Matrix product of ECW_ES_CUTOFFS and W, ie. weighted factors in polynomials
        WC = np.matmul(W, MetabolicPowerModel.ECW_POLY_COEFF)

        # Calculate vel^4 + vel^3 + vel^2 + vel + 1 for every frame and player
        V = np.stack(
            (
                np.power(vel, 4),
                np.power(vel, 3),
                np.power(vel, 2),
                vel,
                np.ones(vel.shape),
            ),
            axis=-1,
        )

        # Multiply WC and V. Calculate sum of terms. Multiply with em
        ECW = np.multiply(np.multiply(WC, V).sum(axis=2), em)

        return ECW

    @staticmethod
    def _calc_ecr(es: np.ndarray, em: np.ndarray, eccr: Numeric = 3.6) -> np.ndarray:
        """Calculates Energy cost of running based on formula (3) and (4) from
        Minetti & Parvei (2018).

        Parameters
        ----------
        es: np.array
            Equivalent slope
        em: np.array
            Equivalent mass
        eccr: Numeric
            Energy cost of constant running. Default is set to 3.6
            :math:`\\frac{J}{kg \\cdot m}` according to di Prampero (2018). Can differ
            for different turfs.

        Returns
        -------
        ecr: np.array
            Energy cost of running
        """
        # Cost of negative gradient from Minetti (2018)
        def _cng(es: np.ndarray):
            return -8.34 * es + eccr * np.exp(13 * es)

        # Cost of positive gradient
        def _cpg(es: np.ndarray):
            return 39.5 * es + eccr * np.exp(-4 * es)

        # Energy cost of running. Where es < 0 calculate cost of negative gradient.
        # Where es >= 0 calculate cost of positive gradient.
        ecr = np.piecewise(es, [es < 0, es >= 0], [_cng, _cpg]) * em

        return ecr

    @staticmethod
    def _calc_ecl(
        es: np.ndarray, vel: np.ndarray, em: np.ndarray, eccr: Numeric = 3.6
    ) -> np.ndarray:
        """Calculate Energy cost of locomotion.

        Parameters
        ----------
        es: np.array
            Equivalent slope
        vel: np.array
            Velocity
        em: np.array
            Equivalent mass
        eccr: Numeric
            Energy cost of constant running. Default is set to 3.6
            :math:`\\frac{J}{kg \\cdot m}` according to di Prampero (2018). Can differ
            for different turfs.

        Returns
        -------
        ecl: np.array
            Energy cost of locomotion
        """
        # Check where locomotion is running
        running = MetabolicPowerModel._is_running(vel, es)
        # Calculate energy cost of walking for entire array
        ecl = MetabolicPowerModel._calc_ecw(es, vel, em)
        # Substitute ecw with energy cost of running where locomotion is running
        ecl[running] = MetabolicPowerModel._calc_ecr(es[running], em[running], eccr)

        return ecl

    @staticmethod
    def _calc_metabolic_power(
        es: np.ndarray,
        vel: np.ndarray,
        em: np.ndarray,
        framerate: Numeric,
        eccr: Numeric = 3.6,
    ) -> np.ndarray:
        """Calculates metabolic power as the product of energy cost of locomotion
        and velocity.

        Parameters
        ----------
        es: np.array
            Equivalent slope
        vel: np.array
            Velocity
        em: np.array
            Equivalent mass
        eccr: Numeric
            Energy cost of constant running. Default is set to 3.6
            :math:`\\frac{J}{kg \\cdot m}` according to di Prampero (2018). Can differ
            for different turfs.

        Returns
        -------
        metp: np.array
            Metabolic power
        """
        # Calculate energy cost of locomotion
        ecl = MetabolicPowerModel._calc_ecl(es, vel, em, eccr)
        # Calculate metabolic power as product of ecl and velocity (m/s)
        metp = ecl * vel / framerate

        return metp

    def fit(
        self, xy: XY, difference: str = "central", axis: str = None, eccr: Numeric = 3.6
    ):
        """Fit the model to the given data and calculate metabolic power for every
        player.

        Notes
        -----
        To give appropriate results, unit of coordinates must be in meter.

        Parameters
        ----------
        xy: XY
            Floodlight XY Data object.
        difference: {'central', 'forward}, optional
            The method of differentiation to calculate velocity and acceleration.
            See :func:`~floodlight.models.kinematics.VelocityModel` for further details.
        axis: {None, 'x', 'y'}, optional
                Optional argument that restricts distance calculation to either the x-
                or y-dimension of the data. If set to None (default), distances are
                calculated in both dimensions.
        eccr: Numeric
            Energy cost of constant running. Default is set to 3.6
            :math:`\\frac{J}{kg \\cdot m}` according to di Prampero (2018). Can differ
            for different turfs.
        """
        # Velocity
        velocity_model = VelocityModel()
        velocity_model.fit(xy, difference=difference, axis=axis)
        velocity = velocity_model.velocity()

        # Acceleration
        acceleration_model = AccelerationModel()
        acceleration_model.fit(xy, difference=difference, axis=axis)
        acceleration = acceleration_model.acceleration()

        # Equivalent slope
        equivalent_slope = MetabolicPowerModel._calc_es(
            velocity.property, acceleration.property
        )
        # Equivalent mass
        equivalent_mass = MetabolicPowerModel._calc_em(equivalent_slope)

        # Metabolic power
        metabolic_power = MetabolicPowerModel._calc_metabolic_power(
            equivalent_slope, velocity.property, equivalent_mass, xy.framerate, eccr
        )

        self._metabolic_power_ = PlayerProperty(
            property=metabolic_power,
            name="metabolic_power",
            framerate=xy.framerate,
        )

    @requires_fit
    def metabolic_power(self) -> PlayerProperty:
        """Returns the frame-wise metabolic power as computed by the ``fit()``-method.

        Returns
        -------
        metabolic_power: PlayerProperty
            A Player Property object of shape (T, N), where T is the total number of
            frames and N is the number of players. The columns contain the frame-wise
            metabolic power.
        """
        metabolic_power = self._metabolic_power_
        return metabolic_power

    @requires_fit
    def cumulative_metabolic_power(self) -> PlayerProperty:
        """Returns the cumulative metabolic power.

        Returns
        -------
        metabolic_power: PlayerProperty
            A Player Property object of shape (T, N), where T is the total number of
            frames and N is the number of players. The columns contain the cumulative
            metabolic power calculated by numpy.nancumsum() over axis=0.
        """
        cum_metp = np.nancumsum(self._metabolic_power_.property, axis=0)
        cumulative_metabolic_power = PlayerProperty(
            property=cum_metp,
            name="cumulative_metabolic_power",
            framerate=self._metabolic_power_.framerate,
        )
        return cumulative_metabolic_power

    @requires_fit
    def equivalent_distance(self, eccr: Numeric = 3.6) -> PlayerProperty:
        """Returns frame-wise equivalent distance, defined as the distance a player
        could have run if moving at a constant speed and calculated as the fraction of
        metabolic work and the cost of constant running.

        Parameters
        ----------
        eccr: Numeric
            Energy cost of constant running. Default is set to 3.6
            :math:`\\frac{J}{kg \\cdot m}` according to di Prampero (2018). Can differ
            for different turfs.

        Returns
        -------
        equivalent_distance: PlayerProperty
            A Player Property object of shape (T, N), where T is the total number of
            frames and N is the number of players. The columns contain the frame-wise
            equivalent distance.
        """
        eq_dist = self._metabolic_power_.property / eccr
        cumulative_metabolic_power = PlayerProperty(
            property=eq_dist,
            name="equivalent_distance",
            framerate=self._metabolic_power_.framerate,
        )
        return cumulative_metabolic_power

    @requires_fit
    def cumulative_equivalent_distance(self, eccr: Numeric = 3.6) -> PlayerProperty:
        """Returns cumulative equivalent distance defined as the distance a player
        could have run if moving at a constant speed and calculated as the fraction
        of metabolic work and the cost of constant running.

        Parameters
        ----------
        eccr: Numeric
            Energy cost of constant running. Default is set to 3.6
            :math:`\\frac{J}{kg \\cdot m}` according to di Prampero (2018). Can differ
            for different turfs.

        Returns
        -------
        cumulative_equivalent_distance: PlayerProperty
            A Player Property object of shape (T, N), where T is the total number of
            frames and N is the number of players. The columns contain the cumulative
            equivalent distance calculated by numpy.nancumsum() over axis=0.
        """
        cum_metp = np.nancumsum(self._metabolic_power_.property, axis=0)
        cum_eqdist = cum_metp / eccr

        cumulative_equivalent_distance = PlayerProperty(
            property=cum_eqdist,
            name="cumulative_equivalent_distance",
            framerate=self._metabolic_power_.framerate,
        )
        return cumulative_equivalent_distance



from typing import Tuple

import numpy as np
from scipy.spatial.distance import cdist

from floodlight import Pitch, XY
from floodlight.core.property import PlayerProperty
from floodlight.utils.types import Numeric
from floodlight.models.base import BaseModel


class VoronoiModel(BaseModel):
    def __init__(self, pitch):
        super().__init__(pitch)

        # initialize instance attributes
        self.mesh_ = None


    def _create_mesh(
            self,
            xlim: Tuple[Numeric, Numeric],
            ylim: Tuple[Numeric, Numeric],
            step_size: int,
    ):
        # init pitch
        mesh = np.meshgrid()

    def fit(self, xy1: XY, xy2: XY) -> None:
        pass






from typing import Tuple

import matplotlib
from matplotlib.patches import Arc, Wedge, Rectangle

from floodlight.utils.types import Numeric


def plot_handball_pitch(
    xlim: Tuple[Numeric, Numeric],
    ylim: Tuple[Numeric, Numeric],
    unit: str,
    color_scheme: str,
    show_axis_ticks: bool,
    ax: matplotlib.axes,
    **kwargs,
) -> matplotlib.axes:
    """Plots a handball pitch on a given matplotlib.axes.

    Parameters
    ----------
    xlim: Tuple[Numeric, Numeric]
        Limits of pitch boundaries in longitudinal direction. This tuple has the form
        (x_min, x_max) and delimits the length of the pitch (not of any actual data)
        within the coordinate system.
    ylim: Tuple[Numeric, Numeric]
        Limits of pitch boundaries in lateral direction. This tuple has the form
        (y_min, y_max) and delimits the width of the pitch (not of any actual data)
        within the coordinate system.
    unit: str
        The unit in which data is measured along axes. Possible types are
        {'m', 'cm', 'percent'}.
    color_scheme: str
        Color scheme of the plot. One of {'standard', 'bw'}.
    show_axis_ticks: bool
        If set to True, the axis ticks are visible.
    ax: matplotlib.axes
        Axes from matplotlib library on which the handball field is plotted.
    kwargs:
        Optional keyworded arguments {'linewidth', 'zorder', 'scalex', 'scaley'}
        which can be used for the plot functions from matplotlib. The kwargs are
        only passed to all the plot functions of matplotlib.

    Returns
    -------
    axes : matplotlib.axes
        Axes from matplotlib library on which a handball pitch is plotted.
    """

    # kwargs which are used to configure the plot with default values 1 and 0.
    # all the other kwargs will be just passed to all the plot functions.
    linewidth = kwargs.pop("linewidth", 1)
    zorder = kwargs.pop("zorder", 0)

    # customizing visualization
    if color_scheme == "bw":
        ax.set_facecolor("white")  # color of the pitch
        color_contour_lines = "black"
        color_goal_area_fill = "white"
        color_goal_posts = "black"
    else:
        ax.set_facecolor("skyblue")  # color of the pitch
        color_contour_lines = "white"
        color_goal_area_fill = "khaki"
        color_goal_posts = "white"

    # Positions and ranges on the field which are used for plotting all lines and
    # arcs.
    # All the positions and ranges are scaled based on percentages of the x-range
    # and y-range.

    # key positions for the handball pitch
    xmin, xmax = xlim
    ymin, ymax = ylim
    x_range = xmax - xmin
    y_range = ymax - ymin
    x_half = (xmax + xmin) / 2
    y_half = (ymax + ymin) / 2

    # Positions (x,y) on the handball pitch
    pos_left_side_right_goal_post = (xmin, ymin + y_range * 0.425)
    pos_left_side_left_goal_post = (xmin, ymin + y_range * 0.575)
    pos_right_side_left_goal_post = (xmax, ymin + y_range * 0.425)
    pos_right_side_right_goal_post = (xmax, ymin + y_range * 0.575)
    pos_start_right_side_rect_goal_area = (
        xmax - x_range * 0.15,
        ymin + y_range * 0.425,
    )

    # Ranges on the handball pitch
    y_height_free_throw_arc = y_range * 0.9
    x_width_free_throw_arc = x_range * 0.45
    y_height_goal_area_arc = y_range * 0.6
    x_width_goal_area_arc = x_range * 0.3
    radius_goal_area_arc = x_range * 0.15
    width_goal = y_range * 0.15
    x_height_goal = x_range * 0.025

    # Y positions on the handball pitch
    y_range_center_to_post = width_goal / 2
    y_pos_goal_lower_post = y_half - y_range_center_to_post
    y_pos_goal_upper_post = y_half + y_range_center_to_post
    y_pos_lower_goal_posts = ymin + y_range * 0.425
    y_pos_upper_goal_posts = ymin + y_range * 0.575
    lower_y_pos_4m_line = ymin + (y_range * 0.49625)
    upper_y_pos_4m_line = ymin + (y_range * 0.50375)
    lower_y_pos_7m_line = ymin + (y_range * 0.475)
    upper_y_pos_7m_line = ymin + (y_range * 0.525)

    # X positions on the handball pitch
    x_pos_left_side_goal_area_line_edge = xmin + (x_range * 0.15)
    x_pos_right_side_goal_area_line_edge = xmax - (x_range * 0.15)
    x_pos_left_side_free_throw_line_edge = xmin + (x_range * 0.225)
    x_pos_right_side_free_throw_line_edge = xmax - (x_range * 0.225)
    x_pos_left_side_4m_line = xmin + (x_range * 0.1)
    x_pos_right_side_4m_line = xmax - (x_range * 0.1)
    x_pos_left_side_7m_line = xmin + (x_range * 0.175)
    x_pos_right_side_7m_line = xmax - (x_range * 0.175)

    # angle for the free throw arc changes when unit is 'percent'
    angle = 10 if unit == "percent" else 0

    # margins of the plot
    x_margin = x_range * 0.025
    y_margin = y_range * 0.05

    # set up the boundaries of ax
    ax.set_xlim([xmin - (x_height_goal + x_margin), xmax + (x_height_goal + x_margin)])
    ax.set_ylim([ymin - y_margin, ymax + y_margin])

    # paint handball pitch with all properties
    # main boundaries
    ax.plot(
        [xmin, xmin],
        [ymin, ymax],
        color=color_contour_lines,
        linewidth=linewidth,
        zorder=zorder,
        **kwargs,
    )
    ax.plot(
        [xmax, xmax],
        [ymin, ymax],
        color=color_contour_lines,
        linewidth=linewidth,
        zorder=zorder,
        **kwargs,
    )
    ax.plot(
        [xmin, xmax],
        [ymin, ymin],
        color=color_contour_lines,
        linewidth=linewidth,
        zorder=zorder,
        **kwargs,
    )
    ax.plot(
        [xmin, xmax],
        [ymax, ymax],
        color=color_contour_lines,
        linewidth=linewidth,
        zorder=zorder,
        **kwargs,
    )

    # midline
    ax.plot(
        [x_half, x_half],
        [ymin, ymax],
        color=color_contour_lines,
        zorder=zorder,
        linewidth=linewidth,
        **kwargs,
    )

    # free-throw lines
    # lower left
    ax.add_patch(
        Arc(
            pos_left_side_right_goal_post,
            width=x_width_free_throw_arc,
            height=y_height_free_throw_arc,
            angle=0,
            theta1=290 - angle,
            theta2=360,
            linestyle="dashed",
            linewidth=linewidth,
            color=color_contour_lines,
            zorder=zorder,
            **kwargs,
        )
    )
    # upper left
    ax.add_patch(
        Arc(
            pos_left_side_left_goal_post,
            width=x_width_free_throw_arc,
            height=y_height_free_throw_arc,
            angle=0,
            theta1=0,
            theta2=70 + angle,
            linestyle="dashed",
            linewidth=linewidth,
            color=color_contour_lines,
            zorder=zorder,
            **kwargs,
        )
    )
    # lower right
    ax.add_patch(
        Arc(
            pos_right_side_left_goal_post,
            width=x_width_free_throw_arc,
            height=y_height_free_throw_arc,
            angle=0,
            theta1=180,
            theta2=250 + angle,
            linestyle="dashed",
            linewidth=linewidth,
            color=color_contour_lines,
            zorder=zorder,
            **kwargs,
        )
    )
    # upper right
    ax.add_patch(
        Arc(
            pos_right_side_right_goal_post,
            width=x_width_free_throw_arc,
            height=y_height_free_throw_arc,
            angle=0,
            theta1=110 - angle,
            theta2=180,
            linestyle="dashed",
            linewidth=linewidth,
            color=color_contour_lines,
            zorder=zorder,
            **kwargs,
        )
    )
    # goal area lines
    # lower left
    # Filling the goal area
    if unit != "percent":  # Since wedges can't be scaled
        ax.add_patch(
            Wedge(
                pos_left_side_right_goal_post,
                r=radius_goal_area_arc,
                theta1=270,
                theta2=360,
                linewidth=linewidth,
                color=color_goal_area_fill,
                zorder=zorder,
                **kwargs,
            )
        )
    # goal area line
    ax.add_patch(
        Arc(
            pos_left_side_right_goal_post,
            height=y_height_goal_area_arc,
            width=x_width_goal_area_arc,
            angle=0,
            theta1=270,
            theta2=360,
            linewidth=linewidth,
            color=color_contour_lines,
            zorder=zorder,
            **kwargs,
        )
    )
    # upper left
    # filling the goal area
    if unit != "percent":  # Since wedges can't be scaled
        ax.add_patch(
            Wedge(
                pos_left_side_left_goal_post,
                r=radius_goal_area_arc,
                theta1=0,
                theta2=90,
                linewidth=linewidth,
                color=color_goal_area_fill,
                zorder=zorder,
                **kwargs,
            )
        )
    # goal area line
    ax.add_patch(
        Arc(
            pos_left_side_left_goal_post,
            height=y_height_goal_area_arc,
            width=x_width_goal_area_arc,
            angle=0,
            theta1=0,
            theta2=90,
            linewidth=linewidth,
            color=color_contour_lines,
            zorder=zorder,
            **kwargs,
        )
    )
    # lower right
    # filling the goal area
    if unit != "percent":  # Since wedges can't be scaled
        ax.add_patch(
            Wedge(
                pos_right_side_left_goal_post,
                r=radius_goal_area_arc,
                theta1=180,
                theta2=270,
                linewidth=linewidth,
                color=color_goal_area_fill,
                zorder=zorder,
                **kwargs,
            )
        )
    # goal area line
    ax.add_patch(
        Arc(
            pos_right_side_left_goal_post,
            height=y_height_goal_area_arc,
            width=x_width_goal_area_arc,
            angle=0,
            theta1=180,
            theta2=270,
            linewidth=linewidth,
            color=color_contour_lines,
            zorder=zorder,
            **kwargs,
        )
    )
    # upper right
    # filling the goal area
    if unit != "percent":  # Since wedges can't be scaled
        ax.add_patch(
            Wedge(
                pos_right_side_right_goal_post,
                r=radius_goal_area_arc,
                theta1=90,
                theta2=180,
                linewidth=linewidth,
                color=color_goal_area_fill,
                zorder=zorder,
                **kwargs,
            )
        )
    # goal area line
    ax.add_patch(
        Arc(
            pos_right_side_right_goal_post,
            height=y_height_goal_area_arc,
            width=x_width_goal_area_arc,
            angle=0,
            theta1=90,
            theta2=180,
            linewidth=linewidth,
            color=color_contour_lines,
            zorder=zorder,
            **kwargs,
        )
    )
    # mid left
    # filling the goal area
    if unit != "percent":  # Since wedges can't be scaled the rectangles
        ax.add_patch(  # are also not used to fill the goal area
            Rectangle(
                pos_left_side_right_goal_post,
                width=radius_goal_area_arc,
                height=width_goal,
                color=color_goal_area_fill,
                zorder=zorder,
                **kwargs,
            )
        )
    # mid right
    # filling the goal area
    if unit != "percent":  # Since wedges can't be scaled the rectangles
        ax.add_patch(  # are also not used to fill the goal area
            Rectangle(
                pos_start_right_side_rect_goal_area,
                width=radius_goal_area_arc,
                height=width_goal,
                color=color_goal_area_fill,
                zorder=zorder,
                **kwargs,
            )
        )

    # vertical goal area lines
    ax.plot(
        [x_pos_left_side_goal_area_line_edge, x_pos_left_side_goal_area_line_edge],
        [y_pos_lower_goal_posts, y_pos_upper_goal_posts],
        color=color_contour_lines,
        linewidth=linewidth,
        zorder=zorder,
        **kwargs,
    )
    ax.plot(
        [
            x_pos_right_side_goal_area_line_edge,
            x_pos_right_side_goal_area_line_edge,
        ],
        [y_pos_lower_goal_posts, y_pos_upper_goal_posts],
        color=color_contour_lines,
        linewidth=linewidth,
        zorder=zorder,
        **kwargs,
    )
    # vertical free throw lines
    ax.plot(
        [
            x_pos_left_side_free_throw_line_edge,
            x_pos_left_side_free_throw_line_edge,
        ],
        [y_pos_lower_goal_posts, y_pos_upper_goal_posts],
        color=color_contour_lines,
        linewidth=linewidth,
        linestyle="dashed",
        zorder=zorder,
        **kwargs,
    )
    ax.plot(
        [
            x_pos_right_side_free_throw_line_edge,
            x_pos_right_side_free_throw_line_edge,
        ],
        [y_pos_lower_goal_posts, y_pos_upper_goal_posts],
        color=color_contour_lines,
        linewidth=linewidth,
        linestyle="dashed",
        zorder=zorder,
        **kwargs,
    )

    # 4 m lines
    ax.plot(
        [x_pos_left_side_4m_line, x_pos_left_side_4m_line],
        [lower_y_pos_4m_line, upper_y_pos_4m_line],
        color=color_contour_lines,
        linewidth=linewidth,
        zorder=zorder,
        **kwargs,
    )
    ax.plot(
        [x_pos_right_side_4m_line, x_pos_right_side_4m_line],
        [lower_y_pos_4m_line, upper_y_pos_4m_line],
        color=color_contour_lines,
        linewidth=linewidth,
        zorder=zorder,
        **kwargs,
    )

    # 7 m lines
    ax.plot(
        [x_pos_left_side_7m_line, x_pos_left_side_7m_line],
        [lower_y_pos_7m_line, upper_y_pos_7m_line],
        color=color_contour_lines,
        linewidth=linewidth,
        zorder=zorder,
        **kwargs,
    )
    ax.plot(
        [x_pos_right_side_7m_line, x_pos_right_side_7m_line],
        [lower_y_pos_7m_line, upper_y_pos_7m_line],
        color=color_contour_lines,
        linewidth=linewidth,
        zorder=zorder,
        **kwargs,
    )

    # goals
    # left goal
    ax.plot(
        [xmin - x_height_goal, xmin - x_height_goal],
        [y_pos_goal_lower_post, y_pos_goal_upper_post],
        color=color_goal_posts,
        linewidth=linewidth,
        zorder=zorder,
        **kwargs,
    )
    ax.plot(
        [xmin - x_height_goal, xmin],
        [y_pos_goal_lower_post, y_pos_goal_lower_post],
        color=color_goal_posts,
        linewidth=linewidth,
        zorder=zorder,
        **kwargs,
    )
    ax.plot(
        [xmin - x_height_goal, xmin],
        [y_pos_goal_upper_post, y_pos_goal_upper_post],
        color=color_goal_posts,
        linewidth=linewidth,
        zorder=zorder,
        **kwargs,
    )
    # right goal
    ax.plot(
        [xmax + x_height_goal, xmax + x_height_goal],
        [y_pos_goal_lower_post, y_pos_goal_upper_post],
        color=color_goal_posts,
        linewidth=linewidth,
        zorder=zorder,
        **kwargs,
    )
    ax.plot(
        [xmax + x_height_goal, xmax],
        [y_pos_goal_lower_post, y_pos_goal_lower_post],
        color=color_goal_posts,
        linewidth=linewidth,
        zorder=zorder,
        **kwargs,
    )
    ax.plot(
        [xmax + x_height_goal, xmax],
        [y_pos_goal_upper_post, y_pos_goal_upper_post],
        color=color_goal_posts,
        linewidth=linewidth,
        zorder=zorder,
        **kwargs,
    )
    # baseline
    ax.plot(
        [xmax, xmax],
        [ymin, ymax],
        color=color_contour_lines,
        linewidth=linewidth,
        zorder=zorder,
        **kwargs,
    )
    # baseline
    ax.plot(
        [xmin, xmin],
        [ymin, ymax],
        color=color_contour_lines,
        linewidth=linewidth,
        zorder=zorder,
        **kwargs,
    )

    # remove labels and ticks
    if not show_axis_ticks:
        ax.xaxis.set_major_locator(matplotlib.ticker.NullLocator())
        ax.yaxis.set_major_locator(matplotlib.ticker.NullLocator())

    return ax


def plot_football_pitch(
    xlim: Tuple[Numeric, Numeric],
    ylim: Tuple[Numeric, Numeric],
    length: Numeric,
    width: Numeric,
    unit: str,
    color_scheme: str,
    show_axis_ticks: bool,
    ax: matplotlib.axes,
    **kwargs,
) -> matplotlib.axes:
    """Plots a football pitch on a given matplotlib.axes.

    Parameters
    ----------
    xlim: Tuple[Numeric, Numeric]
        Limits of pitch boundaries in longitudinal direction. This tuple has the form
        (x_min, x_max) and delimits the length of the pitch (not of any actual data)
        within the coordinate system.
    ylim: Tuple[Numeric, Numeric]
        Limits of pitch boundaries in lateral direction. This tuple has the form
        (y_min, y_max) and delimits the width of the pitch (not of any actual data)
        within the coordinate system.
    length: Numeric
        Length of the actual pitch in `unit`.
    width: Numeric, optional
        Width of the actual pitch in `unit`.
    unit: str
        The unit in which data is measured along axes. Possible types are
        {'m', 'cm', 'percent'}.
    color_scheme: str
        Color scheme of the plot. One of {'standard', 'bw'}.
    show_axis_ticks: bool
        If set to True, the axis ticks are visible.
    ax: matplotlib.axes
        Axes from matplotlib library on which the football field is plotted.
    kwargs:
        Optional keyworded arguments {'linewidth', 'zorder', 'scalex', 'scaley'}
        which can be used for the plot functions from matplotlib. The kwargs are
        only passed to all the plot functions of matplotlib.

    Returns
    -------
    axes : matplotlib.axes
        Axes from matplotlib library on which a football pitch is plotted.
    """
    # kwargs which are used to configure the plot with default values 1 and 0.
    # all the other kwargs will be just passed to all the plot functions.
    linewidth = kwargs.pop("linewidth", 1)
    zorder = kwargs.pop("zorder", 0)

    # customizing visualization
    if color_scheme == "bw":
        ax.set_facecolor("white")  # color of the pitch
        color_contour_lines = "black"
    else:
        ax.set_facecolor("green")  # color of the pitch
        color_contour_lines = "white"

    # Since football pitch sizes can vary while the elements on the pitch
    # (i.e penalty area) have the same size, the actual elements on the pitch can't
    # be scaled based on the x- or y-range.
    # Therefore a norm factor for the x and y direction is needed for every element
    # on the pitch that has a fixed size.
    # The norm factor is specified based on the given unit.
    # If the unit is 'm' or 'cm' the ratio between x and y is set to 1
    # (see in the Pitch.plot() method ax.set_aspect(1)) and the norm factos are set
    # to 1 ('m') or 100 ('cm'). But if the unit is 'percent' the ratio between
    # width/length is set to ax.set_aspect(width/length).
    # That means if an element like the goal area, is drawn, it get's rescaled based
    # on the ratio of width and length.

    # norm_factor for all elements on the pitch that are scaled in the x direction
    norm_factor_x = (
        1
        if unit == "m"
        else 100
        if unit == "cm"
        else 100 / length
        if length
        else 100 / 105
    )
    # norm_factor for all elements on the pitch that are scaled in the y direction
    norm_factor_y = (
        1
        if unit == "m"
        else 100
        if unit == "cm"
        else 100 / width
        if width
        else 100 / 68
    )

    # All the positions and ranges of certain elements on the pitch
    # (i.e the penalty area) have fixed sizes and are scaled based on the
    # x and y norm factors.

    # key positions for the football pitch
    xmin, xmax = xlim
    ymin, ymax = ylim
    x_half = (xmax + xmin) / 2
    y_half = (ymax + ymin) / 2
    x_radius_points = 0.25 * norm_factor_x * 2
    y_radius_points = 0.25 * norm_factor_y * 2
    x_goal_height = 2.44 * norm_factor_x
    y_goal_width = 7.32 * norm_factor_y
    y_range_center_to_post = y_goal_width / 2

    # goal area
    x_pos_left_side_goal_area_line = xmin + 5.5 * norm_factor_x
    x_pos_right_side_goal_area_line = xmax - 5.5 * norm_factor_x
    y_pos_lower_goal_area = y_half - 9.16 * norm_factor_y
    y_pos_upper_goal_area = y_half + 9.16 * norm_factor_y

    # goal
    y_pos_goal_lower_post = y_half - y_range_center_to_post
    y_pos_goal_upper_post = y_half + y_range_center_to_post

    # penalty area
    x_pos_left_side_penalty_area_line = xmin + 16.5 * norm_factor_x
    y_pos_penalty_area_lower_line = y_half - 20.16 * norm_factor_y
    y_pos_penalty_area_upper_line = y_half + 20.16 * norm_factor_y
    x_pos_right_side_penalty_area_line = xmax - 16.5 * norm_factor_x
    x_pos_left_side_center_penalty_arc = xmin + 11 * norm_factor_x
    x_pos_right_side_center_penalty_arc = xmax - 11 * norm_factor_x
    x_pos_left_side_penalty_point = xmin + 11 * norm_factor_x
    x_pos_right_side_penalty_point = xmax - 11 * norm_factor_x

    # center circle
    x_radius_center_circle = 9.15 * norm_factor_x * 2
    y_radius_center_circle = 9.15 * norm_factor_y * 2

    # angle for the penalty area arc changes when unit is 'percent'
    angle = 10 if unit == "percent" else 0

    # field boundaries
    ax.plot(
        [xmin, xmax],
        [ymin, ymin],
        color=color_contour_lines,
        zorder=zorder,
        linewidth=linewidth,
    )
    ax.plot(
        [xmin, xmax],
        [ymax, ymax],
        color=color_contour_lines,
        zorder=zorder,
        linewidth=linewidth,
    )
    ax.plot(
        [xmin, xmin],
        [ymin, ymax],
        color=color_contour_lines,
        zorder=zorder,
        linewidth=linewidth,
    )
    ax.plot(
        [xmax, xmax],
        [ymin, ymax],
        color=color_contour_lines,
        zorder=zorder,
        linewidth=linewidth,
    )
    ax.plot(
        [x_half, x_half],
        [ymin, ymax],
        color=color_contour_lines,
        zorder=zorder,
        linewidth=linewidth,
    )

    # goal area
    # goal area left
    ax.plot(
        [x_pos_left_side_goal_area_line, x_pos_left_side_goal_area_line],
        [y_pos_lower_goal_area, y_pos_upper_goal_area],
        color=color_contour_lines,
        zorder=zorder,
        linewidth=linewidth,
    )
    ax.plot(
        [xmin, x_pos_left_side_goal_area_line],
        [y_pos_lower_goal_area, y_pos_lower_goal_area],
        color=color_contour_lines,
        zorder=zorder,
        linewidth=linewidth,
    )
    ax.plot(
        [xmin, x_pos_left_side_goal_area_line],
        [y_pos_upper_goal_area, y_pos_upper_goal_area],
        color=color_contour_lines,
        zorder=zorder,
        linewidth=linewidth,
    )
    # goal area right
    ax.plot(
        [x_pos_right_side_goal_area_line, x_pos_right_side_goal_area_line],
        [y_pos_lower_goal_area, y_pos_upper_goal_area],
        color=color_contour_lines,
        zorder=zorder,
        linewidth=linewidth,
    )
    ax.plot(
        [xmax, x_pos_right_side_goal_area_line],
        [y_pos_lower_goal_area, y_pos_lower_goal_area],
        color=color_contour_lines,
        zorder=zorder,
        linewidth=linewidth,
    )
    ax.plot(
        [xmax, x_pos_right_side_goal_area_line],
        [y_pos_upper_goal_area, y_pos_upper_goal_area],
        color=color_contour_lines,
        zorder=zorder,
        linewidth=linewidth,
    )

    # penalty area
    # penalty area left
    ax.plot(
        [x_pos_left_side_penalty_area_line, x_pos_left_side_penalty_area_line],
        [y_pos_penalty_area_lower_line, y_pos_penalty_area_upper_line],
        color=color_contour_lines,
        zorder=zorder,
        linewidth=linewidth,
    )
    ax.plot(
        [xmin, x_pos_left_side_penalty_area_line],
        [y_pos_penalty_area_lower_line, y_pos_penalty_area_lower_line],
        color=color_contour_lines,
        zorder=zorder,
        linewidth=linewidth,
    )
    ax.plot(
        [xmin, x_pos_left_side_penalty_area_line],
        [y_pos_penalty_area_upper_line, y_pos_penalty_area_upper_line],
        color=color_contour_lines,
        zorder=zorder,
        linewidth=linewidth,
    )
    # penalty area right
    ax.plot(
        [
            x_pos_right_side_penalty_area_line,
            x_pos_right_side_penalty_area_line,
        ],
        [y_pos_penalty_area_lower_line, y_pos_penalty_area_upper_line],
        color=color_contour_lines,
        zorder=zorder,
        linewidth=linewidth,
    )
    ax.plot(
        [xmax, x_pos_right_side_penalty_area_line],
        [y_pos_penalty_area_lower_line, y_pos_penalty_area_lower_line],
        color=color_contour_lines,
        zorder=zorder,
        linewidth=linewidth,
    )
    ax.plot(
        [xmax, x_pos_right_side_penalty_area_line],
        [y_pos_penalty_area_upper_line, y_pos_penalty_area_upper_line],
        color=color_contour_lines,
        zorder=zorder,
        linewidth=linewidth,
    )
    # penalty point
    # left
    ax.add_patch(
        Arc(
            (x_pos_left_side_penalty_point, y_half),
            width=x_radius_points,
            height=y_radius_points,
            angle=0,
            theta1=0,
            theta2=360,
            color=color_contour_lines,
            zorder=zorder,
            linewidth=linewidth,
        )
    )
    # right
    ax.add_patch(
        Arc(
            (x_pos_right_side_penalty_point, y_half),
            width=x_radius_points,
            height=y_radius_points,
            angle=0,
            theta1=0,
            theta2=360,
            color=color_contour_lines,
            zorder=zorder,
            linewidth=linewidth,
        )
    )
    # penalty area circle left
    ax.add_patch(
        Arc(
            (x_pos_left_side_center_penalty_arc, y_half),
            width=x_radius_center_circle,
            height=y_radius_center_circle,
            angle=0,
            theta1=307.5 - angle,
            theta2=52.5 + angle,
            color=color_contour_lines,
            zorder=zorder,
            linewidth=linewidth,
        )
    )
    # penalty area circle right
    ax.add_patch(
        Arc(
            (x_pos_right_side_center_penalty_arc, y_half),
            width=x_radius_center_circle,
            height=y_radius_center_circle,
            angle=0,
            theta1=127.5 - angle,
            theta2=232.5 + angle,
            color=color_contour_lines,
            zorder=zorder,
            linewidth=linewidth,
        )
    )
    # goal
    # left goal
    ax.plot(
        [xmin - x_goal_height, xmin - x_goal_height],
        [y_pos_goal_lower_post, y_pos_goal_upper_post],
        color=color_contour_lines,
        zorder=zorder,
        linewidth=linewidth,
    )
    ax.plot(
        [xmin - x_goal_height, xmin],
        [y_pos_goal_upper_post, y_pos_goal_upper_post],
        color=color_contour_lines,
        zorder=zorder,
        linewidth=linewidth,
    )
    ax.plot(
        [xmin - x_goal_height, xmin],
        [y_pos_goal_lower_post, y_pos_goal_lower_post],
        color=color_contour_lines,
        zorder=zorder,
        linewidth=linewidth,
    )
    # right goal
    ax.plot(
        [xmax + x_goal_height, xmax + x_goal_height],
        [y_pos_goal_lower_post, y_pos_goal_upper_post],
        color=color_contour_lines,
        zorder=zorder,
        linewidth=linewidth,
    )
    ax.plot(
        [xmax + x_goal_height, xmax],
        [y_pos_goal_upper_post, y_pos_goal_upper_post],
        color=color_contour_lines,
        zorder=zorder,
        linewidth=linewidth,
    )
    ax.plot(
        [xmax + x_goal_height, xmax],
        [y_pos_goal_lower_post, y_pos_goal_lower_post],
        color=color_contour_lines,
        zorder=zorder,
        linewidth=linewidth,
    )

    # center circle
    ax.add_patch(
        Arc(
            (x_half, y_half),
            width=x_radius_center_circle,
            height=y_radius_center_circle,
            angle=0,
            theta1=0,
            theta2=360,
            color=color_contour_lines,
            zorder=zorder,
            linewidth=linewidth,
        )
    )
    # center point
    ax.add_patch(
        Arc(
            (x_half, y_half),
            width=x_radius_points,
            height=y_radius_points,
            angle=0,
            theta1=0,
            theta2=360,
            color=color_contour_lines,
            zorder=zorder,
            linewidth=linewidth,
        )
    )

    # remove labels and ticks
    if not show_axis_ticks:
        ax.xaxis.set_major_locator(matplotlib.ticker.NullLocator())
        ax.yaxis.set_major_locator(matplotlib.ticker.NullLocator())

    return ax



from typing import Tuple

import scipy.signal
import numpy as np

from floodlight import XY
from floodlight.utils.types import Numeric


def _get_filterable_and_short_sequences(
    data: np.ndarray, min_signal_len: int
) -> Tuple[np.ndarray, np.ndarray]:
    """Returns start and end indices of continuous, filterable sequences and sequences
    to short for filtering with the specified filter.

    Parameters
    ----------
    data: np.ndarray
        Array of shape (T,) potentially containing NaNs.
    min_signal_len: int
        The minimum signal length that the specified filter can be applied on.

    Returns
    -------
    filterable_sequences: np.ndarray
        Two-dimensional array of shape (N, 2) and form
        ``[[sequence_start_idx, sequence_end_idx]]`` containing start and end indices of
         N filterable sequences in the original data, ordered ascendingly. A sequence is
         filterable when it doesn't contain NaNs and is at least as long as the minimum
         window length of the specified filter.
    short_sequences: np.ndarray
        Two-dimensional array of shape (N, 2) and form
        ``[[sequence_start_idx, sequence_end_idx]]`` containing start and end indices of
         N sequences in the original data that don't contain NaNs but are to short to
         apply the specified filter on.
    """
    if data.ndim != 1:
        raise ValueError(
            f"Expected input data to be one-dimensional. Got {data.ndim}-dimensional "
            f"data instead."
        )

    # Convert possible None-types in data to np.NaN
    data = np.array(data, dtype=float)

    # indices where nans and numbers are next to each other
    change_points = np.where(np.diff(np.isnan(data), prepend=np.nan, append=np.nan))[0]
    sequences = np.array(
        [
            (change_points[i], change_points[i + 1])
            for i in range(len(change_points) - 1)
        ]
    )

    # which sequences contain NaNs
    seq_is_nan = np.where(np.isnan(data[sequences[:, 0]]), False, True)
    # remove sequences containing NaNs
    non_nan_sequences = sequences[seq_is_nan == 1]
    # split remaining sequences into filterable and short
    filterable_sequences = non_nan_sequences[
        (non_nan_sequences[:, 1] - non_nan_sequences[:, 0]) >= min_signal_len
    ]
    short_sequences = non_nan_sequences[
        (non_nan_sequences[:, 1] - non_nan_sequences[:, 0]) < min_signal_len
    ]

    return filterable_sequences, short_sequences


def _filter_sequence_butterworth_lowpass(
    signal: np.ndarray,
    order: int = 3,
    Wn: Numeric = 1,
    framerate: Numeric = None,
    **kwargs,
) -> np.ndarray:
    """Filters the incoming signal with a digital Butterworth lowpass filter.

    Wrapper for combined application of the `scipy.signal.butter <https://docs.scipy.
    org/doc/scipy/reference/generated/scipy.signal.butter.html>`__ and `scipy.signal.
    filtfilt <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.
    filtfilt.html>`_ functions.

    Parameters
    ----------
    signal: np.ndarray
        Array of shape (T, N) containing the signal to be smoothed with T frames and N
        independent signals. Corresponds to the argument ``x`` from the `scipy.signal.
        filtfilt <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.
        filtfilt.html>`_ function.
    order: int, optional
        The order of the filter. Corresponds to the argument ``N`` from the `scipy.
        signal. butter <https://docs.scipy.org/doc/scipy/reference/generated/scipy.
        signal.butter.html>`_ function. Default is 3.
    Wn: Numeric, optional
        The critical cutoff frequency. Corresponds to the argument ``Wn`` from the
        `scipy.signal.butter <https://docs.scipy.org/doc/scipy/reference/generated/scipy
        .signal.butter.html>`_ function. Default is 1.
    framerate: Numeric, optional
        The sampling frequency of the signal. Corresponds to the argument ``fs`` from
        the `scipy.signal.butter <https://docs.scipy.org/doc/scipy/reference/
        generated/scipy.signal.butter.html>`_ function.
    kwargs:
        Optional arguments {'padtype', 'padlen', 'method', 'irlen'} that can be passed
        to the `scipy.signal.filtfilt <https://docs.scipy.org/doc/scipy/reference/
        generated/scipy.signal.filtfilt.html>`_ function.

    Returns
    -------
    signal_filtered: np.array
        Signal filtered by the Butterworth filter.
    """
    # Calculation of filter coefficients
    coeffs = scipy.signal.butter(
        order,
        Wn,
        btype="lowpass",
        output="ba",
        fs=framerate,
    )
    # applying the filter to the data
    signal_filtered = scipy.signal.filtfilt(
        coeffs[0], coeffs[1], signal, axis=0, **kwargs
    )

    return signal_filtered


def butterworth_lowpass(
    xy: XY, order: int = 3, Wn: Numeric = 1, remove_short_seqs: bool = False, **kwargs
) -> XY:
    """Applies a digital Butterworth lowpass-filter to a XY data object. [1]_

    For filtering, the `scipy.filter.butter <https://docs.scipy.org/doc/scipy/reference/
    generated/scipy.signal.butter.html>`_ and the `scipy.signal.filtfilt <https://docs.
    scipy.org/doc/scipy/reference/generated/scipy.signal.filtfilt.html>`_ functions are
    used. This function provides a convenience access to both functions, directly
    applying the filter to all non-NaN sequences in all columns.

    Parameters
    ----------
    xy: XY
        Floodlight XY Data object.
    order: int, optional
        The order of the filter. Corresponds to the argument ``N`` from the `scipy.
        signal.butter <https://docs.scipy.org/doc/scipy/reference/generated/scipy.
        signal.butter.html>`_ function. Default is 3
    Wn: Numeric, optional
        The critical cutoff frequency. Corresponds to the argument ``Wn`` from the
        `scipy.signal.butter <https://docs.scipy.org/doc/scipy/reference/generated/
        scipy.signal.butter.html>`_ function. Default is 1.
    remove_short_seqs: bool, optional
        If True, sequences that are to short for the filter with the specified settings
        are replaced with np.NaNs. If False, they are kept unfiltered. Default is False.
    kwargs:
        Optional arguments {'padtype', 'padlen', 'method', 'irlen'} that can be passed
        to the `scipy.signal.filtfilt <https://docs.scipy.org/doc/scipy/reference/
        generated/scipy.signal.filtfilt.html>`_ function.
    Returns
    -------
    xy_filtered: XY
        XY object with position data filtered by designed Butterworth low pass filter.

    Notes
    -----
    The values of the input data are assumed to be numerical. Missing data is assumed
    to be either np.NaN or None. The Butterworth-filter requires a minimum signal length
    depending on the settings. A signal is a sequence of data in the XY-object that is
    not interrupted by missing values. The minimum signal length is defined as
    :math:`3 \\cdot (order + 1)`. The treatment of signals shorter than the minimum
    signal length are specified with the ``remove_short_sequence``-argument, where True
    will replace these sequences with np.NaNs ond False will keep the sequences in the
    data unfiltered.

    Examples
    --------
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> from floodlight import XY
    >>> from floodlight.transforms.filter import butterworth_lowpass

    We first generate a noisy XY-object to smooth.

    >>> t = np.linspace(-5, 5, 1000)
    >>> player_x = np.sin(t) * t + np.random.rand(1000)
    >>> player_x[450:495] = np.NaN
    >>> player_x[505:550] = np.NaN
    >>> player_y = t + np.random.randn()
    >>> xy = XY(np.transpose(np.stack((player_x, player_y))), framerate=20)

    Apply the Butterworth lowpass filter with its default settings.

    >>> xy_filt = butterworth_lowpass(xy)
    >>> plt.plot(xy.x)
    >>> plt.plot(xy_filt.x, linewidth=3)
    >>> plt.legend(("Raw", "Smoothed"))
    >>> plt.show()

    .. image:: ../../_img/butterworth_default_example.png


    Apply the same filter but remove the sequence that is to short to filter.

    >>> xy_filt = butterworth_lowpass(xy, remove_short_seqs=True)
    >>> plt.plot(xy.x)
    >>> plt.plot(xy_filt.x, linewidth=3)
    >>> plt.legend(("Raw", "Smoothed"))
    >>> plt.show()

    .. image:: ../../_img/butterworth_removed_short_example.png

    Apply the filter with different specifications.

    >>> xy_filt = butterworth_lowpass(xy, order=5, Wn=4)
    >>> plt.plot(xy.x)
    >>> plt.plot(xy_filt.x, linewidth=3)
    >>> plt.legend(("Raw", "Smoothed"))
    >>> plt.show()

    .. image:: ../../_img/butterworth_adjusted_example.png

    References
    ----------
        .. [1] `Butterworth, S. (1930). On the theory of filter amplifiers. Wireless
            Engineer, 3, 536-541. <https://www.changpuak.ch/electronics/downloads/
            On_the_Theory_of_Filter_Amplifiers.pdf>`_
    """
    # minimum signal length a filter with this specs can be applied on
    min_signal_len = 3 * (order + 1)
    framerate = xy.framerate

    # pre-allocate space for filtered data
    xy_filt = np.empty(xy.xy.shape)
    # loop through the xy-object columns
    for i, column in enumerate(np.transpose(xy.xy)):
        # extract indices of filterable and short sequences
        seqs_filt, seqs_short = _get_filterable_and_short_sequences(
            column, min_signal_len
        )
        # pre-allocate space for filtered column
        col_filt = np.full(column.shape, np.nan)

        # loop through filterable sequences
        for start, end in seqs_filt:
            # apply filter to the sequence and enter filtered data to their
            # respective indices in the data
            col_filt[start:end] = _filter_sequence_butterworth_lowpass(
                column[start:end], order, Wn, framerate, **kwargs
            )
        # check treatment of sequences that don't meet minimum signal length
        if remove_short_seqs is False:
            # enter short sequences unfiltered to their respective indices in the data
            for start, end in seqs_short:
                col_filt[start:end] = column[start:end]

        # enter filtered data into respective column
        xy_filt[:, i] = col_filt

    # create new XY-data object with filtered data
    xy_filtered = XY(xy=xy_filt, framerate=xy.framerate, direction=xy.direction)

    return xy_filtered


def savgol_lowpass(
    xy: XY,
    window_length: int = 5,
    poly_order: Numeric = 3,
    remove_short_seqs: bool = False,
    **kwargs,
) -> XY:
    """Applies a Savitzky-Golay lowpass-filter to a XY data object. [2]_

    For filtering, the `scipy.filter.savgol <https://docs.scipy.org/doc/scipy/reference/
    generated/scipy.signal.savgol_filter.html>`_ function is used. This function
    provides a convenient access to the function, directly applying the filter to all
    non-NaN sequences in all columns.

    Parameters
    ----------
    xy: XY
        Floodlight XY Data object.
    window_length: int, optional
        The length of the filter window. Corresponds to the argument ``window_length``
        from the `scipy.filter.savgol <https://docs.scipy.org/doc/scipy/reference/
        generated/scipy.signal.savgol_filter.html>`_ function. Default is 5.
    polyorder: Numeric, optional
        The order of the polynomial used to fit the samples. ``poly_order`` must be less
        than ``window_length``. Default is 3. Corresponds to the argument ``polyorder``
        from the `scipy.filter.savgol <https://docs.scipy.org/doc/scipy/reference/
        generated/scipy.signal.savgol_filter.html>`_ function. Default is 5.
    remove_short_seqs: bool, optional
        If True, sequences that are to short for the Filter with the specified settings
        are removed from the data. If False, they are kept unfiltered. Default is False.
    kwargs:
        Optional arguments {'deriv', 'delta', 'mode', 'cval'} that can be passed to
        the `scipy.signal.savgol <https://docs.scipy.org/doc/scipy/reference/
        generated/scipy.signal.savgol_filter.html>`_ function.

    Returns
    -------
    xy_filtered: XY
        XY object with position data filtered by designed Butterworth low pass filter.

    Notes
    -----
    The values of the input data are assumed to be numerical. Missing data is assumed
    to be either np.NaN or None. The Savitzky-Golay-filter requires a minimum signal
    length depending on the settings. A signal is a sequence of data in the XY-object
    that is not interrupted by missing values. The minimum signal length is defined as
    the ``window_length``. The treatment of signals shorter than the minimum signal
    length are specified with the ``remove_short_sequence``-argument, where True will
    replace these sequences with np.NaNs ond False will keep the sequences in the data
    unfiltered.

    Examples
    --------
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> from floodlight import XY
    >>> from floodlight.transforms.filter import savgol_lowpass

    We first generate a noisy XY-object to smooth.

    >>> t = np.linspace(-5, 5, 1000)
    >>> player_x = np.sin(t) * t + np.random.rand(1000)
    >>> player_x[450:495] = np.NaN
    >>> player_x[505:550] = np.NaN
    >>> player_y = t + np.random.randn()
    >>> xy = XY(np.transpose(np.stack((player_x, player_y))), framerate=20)

    Apply the Savgol lowpass filter with its default settings.

    >>> xy_filt = savgol_lowpass(xy)
    >>> plt.plot(xy.x)
    >>> plt.plot(xy_filt.x, linewidth=3)
    >>> plt.legend(("Raw", "Smoothed"))
    >>> plt.show()

    .. image:: ../../_img/savgol_default_example.png


    Apply the filter with a longer window lengh and remove the sequence that is to short
    to filter.

    >>> xy_filt = savgol_lowpass(xy, window_length=12, remove_short_seqs=True)
    >>> plt.plot(xy.x)
    >>> plt.plot(xy_filt.x, linewidth=3)
    >>> plt.legend(("Raw", "Smoothed"))
    >>> plt.show()

    .. image:: ../../_img/savgol_removed_short_example.png


    Apply the filter with different specifications.

    >>> xy_filt = savgol_lowpass(xy, window_length=50, poly_order=5)
    >>> plt.plot(xy.x)
    >>> plt.plot(xy_filt.x, linewidth=3)
    >>> plt.legend(("Raw", "Smoothed"))
    >>> plt.show()

    .. image:: ../../_img/savgol_adjusted_example.png

    References
    ----------
        .. [2] `Savitzky, A.; Golay, M.J. (1964). Smoothing and differentiation of data
            by simplified least squares procedures. Analytical Chemistry, 36(1), 1627-
            1639. <https://pubs.acs.org/doi/abs/10.1021/ac60214a047>`_
    """
    # minimum signal length a filter with this specs can be applied on
    min_signal_len = window_length

    # pre-allocate space for filtered data
    xy_filt = np.empty(xy.xy.shape)
    # loop through the xy-object columns
    for i, column in enumerate(np.transpose(xy.xy)):
        # extract indices of alternating NaN/non-NaN sequences
        seqs_filt, seqs_short = _get_filterable_and_short_sequences(
            column, min_signal_len
        )
        # pre-allocate space for filtered column
        col_filt = np.full(column.shape, np.nan)

        # loop through filterable sequences
        for start, end in seqs_filt:
            # apply filter to the sequence and enter filtered data to their
            # respective indices in the data
            col_filt[start:end] = scipy.signal.savgol_filter(
                column[start:end], window_length, poly_order, **kwargs
            )
        # check treatment of sequences that don't meet minimum signal length
        if remove_short_seqs is False:
            # enter short sequences unfiltered to their respective indices in the data
            for start, end in seqs_short:
                col_filt[start:end] = column[start:end]

        # enter filtered data into respective column
        xy_filt[:, i] = col_filt

    # create new XY-data object with filtered data
    xy_filtered = XY(xy=xy_filt, framerate=xy.framerate, direction=xy.direction)

    return xy_filtered



import shutil
import urllib.request
from typing import AnyStr, Any


def extract_zip(filepath: str, target: str, archive_type: str = "zip") -> None:
    """Extracts the content of an archive to disk.

    Parameters
    ----------
    filepath : str
        Path to file.
    target : str
        Target to extract files to.
    archive_type: "zip", optional
        Type of archive, like zip, rar, gzip, etc.
    """
    shutil.unpack_archive(filepath, target, format=archive_type)


def download_from_url(path: str) -> AnyStr:
    """Downloads file from URL.

    Parameters
    ----------
    path : str
        URL path to download data from

    Returns
    -------
    data: AnyStr
    """
    res = urllib.request.urlopen(path)
    return res.read()


def get_and_convert(dic: dict, key: Any, value_type: type, default: Any = None) -> Any:
    """Performs dictionary get and type conversion simultaneously.

    Parameters
    ----------
    dic: dict
        Dictionary to be queried.
    key: Any
        Key to be looked up.
    value_type: type
        Desired output type the value should be cast into.
    default: Any, optional
        Return value if key is not in dic, defaults to None.

    Returns
    -------
    value: value_type
        Returns the value for key if key is in dic, else default. Tries type conversion
        to `type(value) = value_type`. If type conversion fails, e.g. by trying to force
        something like `float(None)` due to a missing dic entry, value is returned in
        its original data type.
    """
    value = dic.get(key, default)
    try:
        value = value_type(value)
    except TypeError:
        pass
    except ValueError:
        pass

    return value



from copy import deepcopy
from dataclasses import dataclass
from typing import Tuple, Union

import matplotlib
import numpy as np

from floodlight.utils.types import Numeric
from floodlight.vis.positions import plot_positions, plot_trajectories


@dataclass
class XY:
    """Spatio-temporal data fragment. Core class of floodlight.

    Attributes
    ----------
    xy: np.ndarray
        Full data array containing x- and y-coordinates, where each player's coordinates
        occupy two consecutive columns.
    x: np.ndarray
        X-data array, where each player's x-coordinates occupy one column.
    y: np.ndarray
        Y-data array, where each player's y-coordinates occupy one column.
    framerate: int, optional
        Temporal resolution of data in frames per second/Hertz.
    direction: str, optional
        Playing direction of players in data fragment, should be either
        'lr' (left-to-right) or 'rl' (right-to-left).
    """

    xy: np.ndarray
    framerate: int = None
    direction: str = None

    def __str__(self):
        return f"Floodlight XY object of shape {self.xy.shape}"

    def __len__(self):
        return len(self.xy)

    def __getitem__(self, key):
        return self.xy[key]

    def __setitem__(self, key, value):
        self.xy[key] = value

    @property
    def x(self):
        return self.xy[:, ::2]

    @x.setter
    def x(self, x_data: np.ndarray):
        self.xy[:, ::2] = x_data

    @property
    def y(self):
        return self.xy[:, 1::2]

    @y.setter
    def y(self, y_data: np.ndarray):
        self.xy[:, 1::2] = y_data

    @property
    def N(self) -> int:
        """Returns the object's number of players."""
        n_columns = self.xy.shape[1]
        if (n_columns % 2) != 0:
            raise ValueError(f"XY has an odd number of columns ({n_columns})")
        return n_columns // 2

    def frame(self, t: int) -> np.ndarray:
        """Returns data for given frame *t*.

        Parameters
        ----------
        t : int
            Frame index.

        Returns
        -------
        frame : np.ndarray
            One-dimensional xy-data row for given frame.
        """
        return self.xy[t, :]

    def player(self, xID: int) -> np.ndarray:
        """Returns data for player with given player index *xID*.

        Parameters
        ----------
        xID : int
            Player index.

        Returns
        -------
        player : np.ndarray
            Two-dimensional xy-data for given player.

        """
        return self.xy[:, xID * 2 : xID * 2 + 2]

    def point(self, t: int, xID: int) -> np.ndarray:
        """Returns data for a point determined by frame *t* and player index *xID*.

        Parameters
        ----------
        t: int
            Frame index.
        xID: int
            Player index.

        Returns
        -------
        point : np.ndarray
            Point-data of shape (2,)
        """
        return self.xy[t, xID * 2 : xID * 2 + 2]

    def translate(self, shift: Tuple[Numeric, Numeric]):
        """Translates data by shift vector.

        Parameters
        ----------
        shift : list or array-like
            Shift vector of form v = (x, y). Any iterable data type with two numeric
            entries is accepted.

        Notes
        -----
        Executing this method will cast the object's xy attribute to dtype np.float32 if
        it previously has a non-floating dtype.
        """
        # cast to float
        if self.xy.dtype not in [np.float_, np.float64, np.float32, float]:
            self.xy = self.xy.astype(np.float32, copy=False)

        self.x = np.round(self.x + shift[0], 3)
        self.y = np.round(self.y + shift[1], 3)

    def scale(self, factor: float, axis: str = None):
        """Scales data by a given factor and optionally selected axis.

        Parameters
        ----------
        factor : float
            Scaling factor.
        axis : {None, 'x', 'y'}, optional
            Name of scaling axis. If set to 'x' data is scaled on x-axis, if set to 'y'
            data is scaled on y-axis. If None, data is scaled in both directions
            (default).

        Notes
        -----
        Executing this method will cast the object's xy attribute to dtype np.float32 if
        it previously has a non-floating dtype.
        """
        # cast to float
        if self.xy.dtype not in [np.float_, np.float64, np.float32, float]:
            self.xy = self.xy.astype(np.float32, copy=False)

        if axis is None:
            self.xy = np.round(self.xy * factor, 3)
        elif axis == "x":
            self.x = np.round(self.x * factor, 3)
        elif axis == "y":
            self.y = np.round(self.y * factor, 3)
        else:
            raise ValueError(f"Expected axis to be one of ('x', 'y', None), got {axis}")

    def reflect(self, axis: str):
        """Reflects data on given `axis`.

        Parameters
        ----------
        axis : {'x', 'y'}
            Name of reflection axis. If set to "x", data is reflected on x-axis,
            if set to "y", data is reflected on y-axis.
        """
        if axis == "x":
            self.scale(factor=-1, axis="y")
        elif axis == "y":
            self.scale(factor=-1, axis="x")
        else:
            raise ValueError(f"Expected axis to be one of ('x', 'y'), got {axis}")

    def rotate(self, alpha: float):
        """Rotates data on given angle 'alpha' around the origin.

        Parameters
        ----------
        alpha: float
            Rotation angle in degrees. Alpha must be between -360 and 360. If positive
            alpha, data is rotated in counter clockwise direction around the origin. If
            negative, data is rotated in clockwise direction around the origin.

        Notes
        -----
        Executing this method will cast the object's xy attribute to dtype np.float32 if
        it previously has a non-floating dtype.
        """
        if not (-360 <= alpha <= 360):
            raise ValueError(
                f"Expected alpha to be from -360 to 360, got {alpha} instead"
            )
        # cast to float
        if self.xy.dtype not in [np.float_, np.float64, np.float32, float]:
            self.xy = self.xy.astype(np.float32, copy=False)

        # construct rotation matrix
        phi = np.radians(alpha)
        cos = np.cos(phi)
        sin = np.sin(phi)
        r = np.array([[cos, -sin], [sin, cos]]).transpose()

        # perform player-wise rotation - this correctly handles nan's compared to
        # block matrix approach
        for p in range(self.N):
            columns = (p * 2, p * 2 + 1)
            self.xy[:, columns] = np.round(self.xy[:, columns] @ r, 3)

    def slice(
        self, startframe: int = None, endframe: int = None, inplace: bool = False
    ):
        """Return copy of object with sliced data. Mimics numpy's array slicing.

        Parameters
        ----------
        startframe : int, optional
            Start of slice. Defaults to beginning of segment.
        endframe : int, optional
            End of slice (endframe is excluded). Defaults to end of segment.
        inplace: bool, optional
            If set to ``False`` (default), a new object is returned, otherwise the
            operation is performed in place on the called object.

        Returns
        -------
        xy_sliced: Union[XY, None]
        """
        sliced_data = self.xy[startframe:endframe, :].copy()
        xy_sliced = None

        if inplace:
            self.xy = sliced_data
        else:
            xy_sliced = XY(
                xy=sliced_data,
                framerate=deepcopy(self.framerate),
                direction=deepcopy(self.direction),
            )

        return xy_sliced

    def plot(
        self,
        t: Union[int, Tuple[int, int]],
        plot_type: str = "positions",
        ball: bool = False,
        ax: matplotlib.axes = None,
        **kwargs,
    ) -> matplotlib.axes:
        """Plots a snapshot or time intervall of the object's spatiotemporal data on a
        matplotlib axes.

        Parameters
        ----------
        t: Union[int, Tuple [int, int]]
            Frame for which postions should be plotted if plot_type == 'positions', or a
            Tuple that has the form (start_frame, end_frame) if
            plot_type == 'trajectories'.
        plot_type: str, optional
            One of {'positions', 'trajectories'}. Determines which plotting function is
            called. Defaults to 'positions'.
        ball: bool, optional
            Boolean indicating whether this object is storing ball data. If set to True,
            the styling is adjusted accordingly. Defaults to False.
        ax: matplotlib.axes, optional
            Axes from matplotlib library to plot on. Defaults to None.
        kwargs:
            Optional keyworded arguments e.g. {'color', 'zorder', 'marker', 'linestyle',
            'alpha'} which can be used for the plot functions from matplotlib.
            The kwargs are only passed to all the plot functions of matplotlib. If not
            given default values are used (see floodlight.vis.positions).


        Returns
        -------
        axes: matplotlib.axes
            Axes from matplotlib library on which the specified plot type is plotted.

        Notes
        -----
        The kwargs are only passed to the plot functions of matplotlib. To customize the
        plots have a look at
        `matplotlib
        <https://matplotlib.org/3.5.0/api/_as_gen/matplotlib.axes.Axes.plot.html>`_.
        For example in order to modify the color of the points and lines pass a color
        name or rgb-value (`matplotlib colors
        <https://matplotlib.org/3.5.0/tutorials/colors/colors.html>`_) to the keyworded
        argument 'color'. The same principle applies to other kwargs like 'zorder',
        'marker' and 'linestyle'.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> import numpy as np
        >>> from floodlight.core.pitch import Pitch
        >>> from floodlight.core.xy import XY
        >>> # positions
        >>> pos = np.array(
        >>>     [[35,5,35,63,25,25,25,50],
        >>>     [45,10,45,55,35,20,35,45],
        >>>     [55,10,55,55,45,20,45,45],
        >>>     [88.5,20,88.5,30,88.5,40,88.5,50]])
        >>> xy_pos = XY(pos) # create XY object
        >>> # create Pitch object
        >>> football_pitch = Pitch(xlim=(0,105), ylim=(0, 68), unit="m",
        >>> sport="football")
        >>> # create matplotlib.axes
        >>> ax = plt.subplots()[1]
        >>> # plot pitch on ax
        >>> football_pitch.plot(color_scheme="standard", ax=ax)

        >>> # plot positions on ax
        >>> xy_pos.plot(plot_type="positions", t=0, ax=ax)
        >>> plt.show()

        .. image:: ../../_img/positions_example.png

        >>> # plot trajectories from frame 0 to 4 on ax
        >>> xy_pos.plot(plot_type="trajectories", t=(0,4), ball= False, ax=ax)
        >>> plt.show()

        .. image:: ../../_img/trajectories_example.png

        """

        plot_types = ["positions", "trajectories"]

        # call visualization function based on plot_type
        if plot_type == "positions":
            return plot_positions(self, t, ball, ax=ax, **kwargs)
        elif plot_type == "trajectories":
            return plot_trajectories(self, t[0], t[1], ball, ax=ax, **kwargs)
        else:
            raise ValueError(
                f"Expected plot_type to be one of {plot_types}, got {plot_type} "
                "instead."
            )



import os.path
import warnings
from typing import Dict, Tuple, Union
from pathlib import Path

import numpy as np
import pandas as pd
from lxml import etree

from floodlight.io.utils import download_from_url, get_and_convert
from floodlight.core.code import Code
from floodlight.core.events import Events
from floodlight.core.pitch import Pitch
from floodlight.core.xy import XY
from settings import DATA_DIR

# ----------------------------- StatsPerform Open Format -------------------------------


def _create_metadata_from_open_csv_df(
    csv_df: pd.DataFrame,
) -> Tuple[Dict[int, tuple], Pitch]:
    """Creates meta information from a pd.DataFrame that results from parsing the open
    StatsPerform event data csv file.

    Parameters
    ----------
    csv_df: pd.DataFrame
        Data Frame with the parsed event data csv file.

    Returns
    -------
    periods: Dict[int, int]
        Dictionary with start and endframes:
            ``periods[segment] = (startframe, endframe)``.
    pitch: Pitch
        Playing Pitch object.
    """

    # create pitch
    pi_len = csv_df["pitch_dimension_long_side"].values[0]
    pi_wid = csv_df["pitch_dimension_short_side"].values[0]
    pitch = Pitch.from_template(
        "statsperform_open",
        length=pi_len,
        width=pi_wid,
        sport="football",
    )

    # create periods for segments, coded as jumps in the frame sequence
    periods = {}
    frame_values = csv_df["frame_count"].unique()

    seg_idx = np.where(np.diff(frame_values, prepend=frame_values[0]) > 1)
    seg_idx = np.insert(seg_idx, 0, 0)
    seg_idx = np.append(seg_idx, len(frame_values))
    for segment in range(len(seg_idx) - 1):
        start = int(frame_values[seg_idx[segment]])
        end = int(frame_values[seg_idx[segment + 1] - 1])
        periods[segment] = (start, end)

    return periods, pitch


def _create_links_from_open_csv_df(
    csv_df: pd.DataFrame, team_ids: Dict[str, float]
) -> Dict[str, Dict[int, int]]:
    """Checks the entire parsed open StatsPerform event data csv file for unique jIDs
    (jerseynumbers) and creates a dictionary linking jIDs to xIDs in ascending order.

    Parameters
    ----------
    csv_df: pd.DataFrame
        Data Frame with the parsed event data csv file.
    team_ids: Dict[str, float]
        Dictionary that stores the StatsPerform team id of the Home and Away team

    Returns
    -------
    links: Dict[str, Dict[int, int]]
        A link dictionary of the form ``links[team][jID] = xID``.
    """
    links = {}
    for team in team_ids:
        links[team] = {
            int(jID): xID
            for xID, jID in enumerate(
                csv_df[csv_df["team_id"] == team_ids[team]]["jersey_no"].unique()
            )
        }
    return links


def _read_open_event_csv_single_line(
    line: str,
) -> Tuple[Dict, str, str]:
    """Extracts all relevant information from a single line of StatsPerform's Event csv
    file (i.e. one single event in the data).

    Parameters
    ----------
    line: str
        One full line from StatsPerform's Event csv file.

    Returns
    -------
    event: Dict
        Dictionary with relevant event information in the form:
        ``event[attribute] = value``.
    """
    event = {}
    attrib = line.split(sep=",")

    # description
    event["eID"] = attrib[5].replace(" ", "")

    # relative time
    event["gameclock"] = float(attrib[4])
    event["frameclock"] = float(attrib[2])

    # segment, player and team
    segment = attrib[3]
    team = attrib[9]
    event["tID"] = team
    event["pID"] = attrib[8]

    # outcome
    event["outcome"] = np.nan
    if "Won" in attrib[5].split(" "):
        event["outcome"] = 1
    elif "Lost" in attrib[5].split(" "):
        event["outcome"] = 0

    # minute and second of game
    event["minute"] = np.floor(event["gameclock"] / 60)
    event["second"] = np.floor(event["gameclock"] - event["minute"] * 60)

    # additional information (qualifier)
    event["qualifier"] = {
        "event_id": attrib[1],
        "event_type_id": attrib[6],
        "sequencenumber": attrib[7],
        "jersey_no": attrib[10],
        "is_pass": attrib[11],
        "is_cross": attrib[12],
        "is_corner": attrib[13],
        "is_free_kick": attrib[14],
        "is_goal_kick": attrib[15],
        "passtypeid": attrib[16],
        "wintypeid": attrib[17],
        "savetypeid": attrib[18],
        "possessionnumber": attrib[19],
    }

    return event, team, segment


def create_links_from_open_tracking_data_csv(
    filepath_tracking: Union[str, Path]
) -> Dict[str, Dict[int, int]]:
    """Parses the entire open StatsPerform event data csv file for unique jIDs
    (jerseynumbers) and creates a dictionary linking jIDs to xIDs in ascending order.

    Parameters
    ----------
    filepath_tracking: str or pathlib.Path
        csv file where the position data in StatsPerform format is saved.

    Returns
    -------
    links: Dict[str, Dict[int, int]]
        A link dictionary of the form ``links[team][jID] = xID``.
    """
    # read dat-file into pd.DataFrame
    dat_df = pd.read_csv(str(filepath_tracking))

    # initialize team and ball ids
    team_ids = {"Home": 1.0, "Away": 2.0}
    ball_id = 4

    # check for additional tIDs
    for ID in dat_df["team_id"].unique():
        if not (ID in team_ids.values() or ID == ball_id):
            warnings.warn(f"Team ID {ID} did not match any of the standard IDs!")

    return _create_links_from_open_csv_df(dat_df, team_ids)


def read_open_event_data_csv(
    filepath_events: Union[str, Path],
) -> Tuple[Events, Events, Events, Events]:
    """Parses an open StatsPerform Match Event csv file and extracts the event data.

    This function provides a high-level access to the particular openly published
    StatsPerform match events csv file (e.g. for the Pro Forum '22) and returns Event
    objects for both teams.

    Parameters
    ----------
    filepath_events: str or pathlib.Path
        Full path to xml File where the Event data in StatsPerform csv format is
        saved
    Returns
    -------
    data_objects: Tuple[Events, Events, Events, Events]
        Events-objects for both teams and both halves.

    Notes
    -----
    StatsPerform's open format of handling provides certain additional event attributes,
    which attach additional information to certain events. As of now, these information
    are parsed as a string in the ``qualifier`` column of the returned DataFrame and can
    be transformed to a dict of form ``{attribute: value}``.
    """
    # initialize bin and variables
    events = {}
    teams = ["1.0", "2.0"]
    segments = ["1", "2"]
    for team in teams:
        events[team] = {segment: pd.DataFrame() for segment in segments}

    # parse event data
    with open(str(filepath_events), "r") as f:
        while True:
            line = f.readline()

            # terminate if at end of file
            if len(line) == 0:
                break

            # skip the head
            if line.split(sep=",")[3] == "current_phase":
                continue

            # read single line
            event, team, segment = _read_open_event_csv_single_line(line)

            # insert to bin
            if team:
                events[team][segment] = events[team][segment].append(
                    event, ignore_index=True
                )
            else:  # if no clear assignment possible, insert to bins for both teams
                for team in teams:
                    events[team][segment] = events[team][segment].append(
                        event, ignore_index=True
                    )

    # assembly
    t1_ht1 = Events(
        events=events["1.0"]["1"],
    )
    t1_ht2 = Events(
        events=events["1.0"]["2"],
    )
    t2_ht1 = Events(
        events=events["2.0"]["1"],
    )
    t2_ht2 = Events(
        events=events["2.0"]["2"],
    )
    data_objects = (t1_ht1, t1_ht2, t2_ht1, t2_ht2)

    return data_objects


def read_open_tracking_data_csv(
    filepath_tracking: Union[str, Path],
    links: Dict[str, Dict[int, int]] = None,
) -> Tuple[XY, XY, XY, XY, XY, XY, Code, Code, Pitch]:
    """Parses an open StatsPerform csv file and extract position data and possession
    codes as well as pitch information.

     Openly published StatsPerform position data (e.g. for the Pro Forum '22) is stored
     in a csv file containing all position data (for both halves) as well as information
     about players, the pitch, and the ball possession. This function provides a
     high-level access to StatsPerform data by parsing the csv file.

    Parameters
    ----------
    filepath_tracking: str or pathlib.Path
        Full path to the csv file.
    links: Dict[str, Dict[int, int]], optional
        A link dictionary of the form ``links[team][jID] = xID``. Player's are
        identified in StatsPerform files via jID, and this dictionary is used to map
        them to a specific xID in the respective XY objects. Should be supplied if that
        order matters. If None is given (default), the links are automatically extracted
        from the csv file at the cost of a second pass through the entire file.

    Returns
    -------
    data_objects: Tuple[XY, XY, XY, XY, XY, XY, Code, Code, Pitch]
        XY-, Code-, and Pitch-objects for both teams and both halves. The order is
        (home_ht1, home_ht2, away_ht1, away_ht2, ball_ht1, ball_ht2,
        possession_ht1, possession_ht2, pitch)
    """
    # parse the csv file into pd.DataFrame
    dat_df = pd.read_csv(str(filepath_tracking))

    # initialize team and ball ids
    team_ids = {"Home": 1.0, "Away": 2.0}
    ball_id = 4

    # check for additional tIDs
    for ID in dat_df["team_id"].unique():
        if not (ID in team_ids.values() or ID == ball_id):
            warnings.warn(f"Team ID {ID} did not match any of the standard IDs!")

    # create or check links
    if links is None:
        links = _create_links_from_open_csv_df(dat_df, team_ids)
    else:
        pass
        # potential check vs jerseys in dat file

    # create periods and pitch
    periods, pitch = _create_metadata_from_open_csv_df(dat_df)
    segments = list(periods.keys())

    # infer data shapes
    number_of_players = {team: len(links[team]) for team in links}
    number_of_frames = {}
    for segment in segments:
        start = periods[segment][0]
        end = periods[segment][1]
        number_of_frames[segment] = end - start + 1

    # bins
    codes = {"possession": {segment: [] for segment in segments}}
    xydata = {
        "Home": {
            segment: np.full(
                [
                    number_of_frames[segment],
                    number_of_players[list(links.keys())[0]] * 2,
                ],
                np.nan,
            )
            for segment in periods
        },
        "Away": {
            segment: np.full(
                [
                    number_of_frames[segment],
                    number_of_players[list(links.keys())[1]] * 2,
                ],
                np.nan,
            )
            for segment in periods
        },
        "Ball": {
            segment: np.full([number_of_frames[segment], 2], np.nan)
            for segment in periods
        },
    }

    # loop
    for segment in segments:

        # teams
        for team in team_ids:
            team_df = dat_df[dat_df["team_id"] == team_ids[team]]
            for pID in team_df["player_id"].unique():
                # extract player information
                pl_df = team_df[team_df["player_id"] == pID]
                frames = pl_df["frame_count"].values
                x_position = pl_df["pos_x"].values
                y_position = pl_df["pos_y"].values

                # compute appearance of player in segment
                appearance = np.array(
                    [
                        (periods[segment][0] <= frame <= periods[segment][-1])
                        for frame in frames
                    ]
                )
                # check for players that did not play in segment
                if not np.sum(appearance):
                    continue

                # insert player position to bin array
                jrsy = int(pl_df["jersey_no"].values[0])
                x_col = (links[team][jrsy] - 1) * 2
                y_col = (links[team][jrsy] - 1) * 2 + 1
                start = frames[appearance][0] - periods[segment][0]
                end = frames[appearance][-1] - periods[segment][0] + 1
                xydata[team][segment][start:end, x_col] = x_position[appearance]
                xydata[team][segment][start:end, y_col] = y_position[appearance]

        # ball
        ball_df = dat_df[dat_df["team_id"] == 4]
        frames = ball_df["frame_count"].values
        appearance = np.array(
            [(periods[segment][0] <= frame <= periods[segment][-1]) for frame in frames]
        )
        xydata["Ball"][segment][:, 0] = ball_df["pos_x"].values[appearance]
        xydata["Ball"][segment][:, 1] = ball_df["pos_x"].values[appearance]

        # update codes
        codes["possession"][segment] = ball_df["possession"].values[appearance]

    # create XY objects
    home_ht1 = XY(xy=xydata["Home"][0], framerate=10)
    home_ht2 = XY(xy=xydata["Home"][1], framerate=10)
    away_ht1 = XY(xy=xydata["Away"][0], framerate=10)
    away_ht2 = XY(xy=xydata["Away"][1], framerate=10)
    ball_ht1 = XY(xy=xydata["Ball"][0], framerate=10)
    ball_ht2 = XY(xy=xydata["Ball"][1], framerate=10)

    # create Code objects
    poss_ht1 = Code(
        name="possession",
        code=codes["possession"][0],
        definitions=dict([(team_id, team) for team, team_id in team_ids.items()]),
        framerate=10,
    )
    poss_ht2 = Code(
        name="possession",
        code=codes["possession"][1],
        definitions=dict([(team_id, team) for team, team_id in team_ids.items()]),
        framerate=10,
    )

    data_objects = (
        home_ht1,
        home_ht2,
        away_ht1,
        away_ht2,
        ball_ht1,
        ball_ht2,
        poss_ht1,
        poss_ht2,
        pitch,
    )
    return data_objects


# ----------------------------- StatsPerform Format ---------------------------


def _read_tracking_data_txt_single_line(
    line: str,
) -> Tuple[
    int,
    int,
    Dict[str, Dict[str, Tuple[float, float, float]]],
    Dict[str, Union[str, tuple]],
]:
    """Extracts all relevant information from a single line of StatsPerform's tracking
    data .txt file (i.e. one frame of data).

    Parameters
    ----------
    line: str
        One full line from StatsPerform's .txt-file, equals one sample of data.

    Returns
    -------
    gameclock: int
        The gameclock of the current segment in milliseconds.
    segment: int
        The segment identifier.
    positions: Dict[str, Dict[str, Tuple[float, float, float]]]
        Nested dictionary that stores player position information for each team and
        player. Has the form ``positions[team][jID] = (x, y)``.
    ball: Dict[str]
        Dictionary with ball information. Has keys 'position', 'possession' and
        'ballstatus'.
    """
    # bins
    positions = {"Home": {}, "Away": {}, "Other": {}}
    ball = {}

    # read chunks
    chunks = line.split(":")
    time_chunk = chunks[0]
    player_chunks = chunks[1].split(";")

    ball_chunk = None
    if len(chunks) > 2:  # check if ball information exist in chunk
        ball_chunk = chunks[2]

    # time chunk
    # systemclock = time_chunk.split(";")[0]
    # possible check or synchronization step
    timeinfo = time_chunk.split(";")[1].split(",")
    gameclock = int(timeinfo[0])
    segment = int(timeinfo[1])
    # ballstatus = timeinfo[2].split(":")[0] == '0'  # '0' seems to be always the case?

    # player chunks
    for player_chunk in player_chunks:

        # skip final entry of chunk
        if not player_chunk or player_chunk == "\n":
            continue

        # read team
        chunk_data = player_chunk.split(",")
        if chunk_data[0] in ["0", "3"]:
            team = "Home"
        elif chunk_data[0] in ["1", "4"]:
            team = "Away"
        else:
            team = "Other"

        # read IDs
        # pID = chunk_data[1]
        jID = chunk_data[2]

        # read positions
        x, y = map(lambda x: float(x), chunk_data[3:])

        # assign
        positions[team][jID] = (x, y)

    # ball chunk
    if ball_chunk is not None:
        x, y, z = map(lambda x: float(x), ball_chunk.split(";")[0].split(","))
        # ball["position"] = (x, y, z)  # z-coordinate is not yet supported
        ball["position"] = (x, y)

    return gameclock, segment, positions, ball


def _read_time_information_from_tracking_data_txt(
    filepath_txt: Union[str, Path],
) -> Tuple[Dict, Union[int, None]]:
    """Reads StatsPerform's tracking .txt file and extracts information about the first
    and last frame of periods. Also, a framerate is estimated from the
    gameclock difference between samples.

    Parameters
    ----------
    filepath_txt: str or pathlib.Path
        Full path to the txt file containing the tracking data.

    Returns
    -------
    periods: Dict
        Dictionary with start and endframes:
        ``periods[segment] = [startframe, endframe]``.
    framerate_est: int or None
        Estimated temporal resolution of data in frames per second/Hertz.
    """

    # bins
    startframes = {}
    endframes = {}
    framerate_est = None

    # read txt file from disk
    file_txt = open(filepath_txt, "r")

    # loop
    last_gameclock = None
    last_segment = None
    for line in file_txt.readlines():

        # read gameclock and segment
        gameclock, segment, _, _ = _read_tracking_data_txt_single_line(line)

        # update periods
        if segment not in startframes:
            startframes[segment] = gameclock
            if last_gameclock is not None:
                endframes[last_segment] = last_gameclock

        # estimate framerate if desired
        if last_gameclock is not None:
            delta = np.absolute(gameclock - last_gameclock)  # in milliseconds
            if framerate_est is None:
                framerate_est = int(1000 / delta)
            elif framerate_est != int(1000 / delta) and last_segment == segment:
                warnings.warn(
                    f"Framerate estimation yielded diverging results."
                    f"The originally estimated framerate of {framerate_est} Hz did not "
                    f"match the current estimation of {int(1000 / delta)} Hz. This "
                    f"might be caused by missing frame(s) in the position data. "
                    f"Continuing by choosing the latest estimation of "
                    f"{int(1000 / delta)} Hz"
                )
                framerate_est = int(1000 / delta)

        # update variables
        last_gameclock = gameclock
        last_segment = segment

    # update end of final segment
    endframes[last_segment] = last_gameclock

    # assembly
    periods = {
        segment: (startframes[segment], endframes[segment]) for segment in startframes
    }

    # close file
    file_txt.close()

    return periods, framerate_est


def _read_jersey_numbers_from_tracking_data_txt(
    file_location_txt: Union[str, Path],
) -> Tuple[set, set]:
    """Reads StatsPerform's tracking .txt file and extracts unique set of jIDs
    (jerseynumbers) for both teams.

    Parameters
    ----------
    file_location_txt: str or pathlib.Path
        Full path to the txt file containing the tracking data.

    Returns
    -------
    home_jIDs: set
    away_jIDs: set
    """

    # bins
    home_jIDs = set()
    away_jIDs = set()

    # read txt file from disk
    file_txt = open(file_location_txt, "r")

    # loop
    for package in file_txt.readlines():

        # read line
        _, _, positions, _ = _read_tracking_data_txt_single_line(package)

        # extract jersey numbers
        home_jIDs |= set(positions["Home"].keys())
        away_jIDs |= set(positions["Away"].keys())

    # close file
    file_txt.close()

    return home_jIDs, away_jIDs


def create_links_from_statsperform_tracking_data_txt(
    filepath_txt: Union[str, Path],
) -> Dict[str, Dict[int, int]]:
    """Parses the entire StatsPerform .txt file for unique jIDs (jerseynumbers) and
    creates a dictionary linking jIDs to xIDs in ascending order.

    Parameters
    ----------
    filepath_txt: str or pathlib.Path
        Full path to the txt file containing the tracking data.

    Returns
    -------
    links: Dict[str, Dict[int, int]]
        Link-dictionary of the form ``links[team][jID] = xID``.
    """
    homejrsy, awayjrsy = _read_jersey_numbers_from_tracking_data_txt(filepath_txt)

    homejrsy = list(homejrsy)
    awayjrsy = list(awayjrsy)

    homejrsy.sort()
    awayjrsy.sort()

    links = {
        "Home": {jID: xID for xID, jID in enumerate(homejrsy)},
        "Away": {jID: xID for xID, jID in enumerate(awayjrsy)},
    }

    return links


def read_event_data_xml(
    filepath_xml: Union[str, Path],
) -> Tuple[Events, Events, Events, Events, Pitch]:
    """Parses a StatsPerform .xml file and extracts event data and pitch information.

    This function provides a high-level access to the StatsPerform match events xml file
    and returns Events objects for both teams and information about the pitch.

    Parameters
    ----------
    filepath_xml: str or pathlib.Path
        Full path to the xml file containing the event data.

    Returns
    -------
    data_objects: Tuple[Events, Events, Events, Events, Pitch]
        Events-objects for both teams and both halves and pitch information. The order
        is (home_ht1, home_ht2, away_ht1, away_ht2, pitch).
    """
    # load xml tree into memory
    tree = etree.parse(str(filepath_xml))
    root = tree.getroot()

    # create bins, read segments, and assign teams
    columns = [
        "eID",
        "gameclock",
        "pID",
        "jID",
        "minute",
        "second",
        "at_x",
        "at_y",
        "to_x",
        "to_y",
        "qualifier",
    ]
    segments = [
        f"HT{get_and_convert(period.attrib, 'IdHalf', str)}"
        for period in root.findall("Events/EventsHalf")
    ]
    teams = ["Home", "Away"]

    # bins
    event_lists = {
        team: {segment: {col: [] for col in columns} for segment in segments}
        for team in teams
    }

    # create links
    links_pID_to_tID = {}
    links_pID_to_jID = {}
    links_pID_to_name = {}
    for teamsheet in root.findall("MatchSheet/Team"):
        # skip referees
        if teamsheet.attrib["Type"] == "Referees":
            continue
        # read team
        team = teamsheet.attrib["Type"][:-4]  # cut 'Team' of e.g. 'HomeTeam'
        # assign player ids to team
        for actor in teamsheet.findall("Actor"):
            if actor.attrib["Occupation"] != "Player":  # coaches etc.
                continue
            links_pID_to_tID[get_and_convert(actor, "IdActor", int)] = team
            links_pID_to_jID[get_and_convert(actor, "IdActor", int)] = get_and_convert(
                actor, "JerseyNumber", int
            )
            links_pID_to_name[get_and_convert(actor, "IdActor", int)] = get_and_convert(
                actor, "NickName", str
            )

    # loop over events
    for half in root.findall("Events/EventsHalf"):
        # get segment information
        period = get_and_convert(half.attrib, "IdHalf", str)
        segment = "HT" + str(period)
        for event in half.findall("Event"):
            # read pID
            pID = get_and_convert(event.attrib, "IdActor1", int)

            # assign team
            team = get_and_convert(links_pID_to_tID, pID, str)

            # create list of either a single team or both teams if no clear assignment
            if team == "None":
                teams_assigned = teams  # add to both teams
            else:
                teams_assigned = [team]  # only add to one team

            # identifier
            eID = get_and_convert(event.attrib, "EventName", str)
            jID = get_and_convert(links_pID_to_jID, pID, int)
            for team in teams_assigned:
                event_lists[team][segment]["eID"].append(eID)
                event_lists[team][segment]["pID"].append(pID)
                event_lists[team][segment]["jID"].append(jID)

            # relative time
            gameclock = get_and_convert(event.attrib, "Time", int)
            minute = np.floor(gameclock / 60)
            second = np.floor(gameclock - minute * 60)
            for team in teams_assigned:
                event_lists[team][segment]["gameclock"].append(gameclock)
                event_lists[team][segment]["minute"].append(minute)
                event_lists[team][segment]["second"].append(second)

            # location
            at_x = get_and_convert(event.attrib, "LocationX", float)
            at_y = get_and_convert(event.attrib, "LocationY", float)
            to_x = get_and_convert(event.attrib, "TargetX", float)
            to_y = get_and_convert(event.attrib, "TargetY", float)
            for team in teams_assigned:
                event_lists[team][segment]["at_x"].append(at_x)
                event_lists[team][segment]["at_y"].append(at_y)
                event_lists[team][segment]["to_x"].append(to_x)
                event_lists[team][segment]["to_y"].append(to_y)

            # qualifier
            qual_dict = {}
            for qual_id in event.attrib:
                qual_value = event.attrib.get(qual_id)
                qual_dict[qual_id] = qual_value
            for team in teams_assigned:
                event_lists[team][segment]["qualifier"].append(str(qual_dict))

    # create pitch
    length = get_and_convert(root.attrib, "FieldLength", int)
    width = get_and_convert(root.attrib, "FieldWidth", int)
    pitch = Pitch.from_template(
        "statsperform",
        length=length,
        width=width,
        sport="football",
    )

    # assembly
    home_ht1 = Events(
        events=pd.DataFrame(data=event_lists["Home"]["HT1"]),
    )
    home_ht2 = Events(
        events=pd.DataFrame(data=event_lists["Home"]["HT2"]),
    )
    away_ht1 = Events(
        events=pd.DataFrame(data=event_lists["Away"]["HT1"]),
    )
    away_ht2 = Events(
        events=pd.DataFrame(data=event_lists["Away"]["HT2"]),
    )

    data_objects = (home_ht1, home_ht2, away_ht1, away_ht2, pitch)

    return data_objects


def read_tracking_data_txt(
    filepath_txt: Union[str, Path],
    links: Dict[str, Dict[int, int]] = None,
) -> Tuple[XY, XY, XY, XY, XY, XY]:
    """Parses a StatsPerform .txt file and extracts position data.

     Internal StatsPerform position data is stored as a .txt file containing all
     position data (for both halves). This function provides a high-level access to
     StatsPerform data by parsing the txt file. Since no information about framerate is
     delivered in the data itself, it is estimated from time difference between
     individual frames.

    Parameters
    ----------
    filepath_txt: str or pathlib.Path
        Full path to the txt file containing the tracking data.
    links: Dict[str, Dict[int, int]], optional
        A link dictionary of the form ``links[team][jID] = xID``. Player's are
        identified in StatsPerform files via jID, and this dictionary is used to map
        them to a specific xID in the respective XY objects. Should be supplied if that
        order matters. If None is given (default), the links are automatically extracted
        from the csv file at the cost of a second pass through the entire file.

    Returns
    -------
    data_objects: Tuple[XY, XY, XY, XY, XY, XY]
        XY-objects for both teams and both halves. The order is (home_ht1, home_ht2,
        away_ht1, away_ht2, ball_ht1, ball_ht2).
    """
    # parse txt file for periods and estimate framerate if not contained in filepath
    periods, framerate_est = _read_time_information_from_tracking_data_txt(filepath_txt)
    segments = list(periods.keys())

    # create or check links
    if links is None:
        links = create_links_from_statsperform_tracking_data_txt(filepath_txt)
    else:
        pass
        # potential check vs jerseys in txt file

    # infer data array shapes
    number_of_home_players = max(links["Home"].values()) + 1
    number_of_away_players = max(links["Away"].values()) + 1
    number_of_frames = {}
    for segment in segments:
        number_of_frames[segment] = (
            int((periods[segment][1] - periods[segment][0]) / 1000 * framerate_est) + 1
        )

    # bins
    xydata = {}
    xydata["Home"] = {
        segment: np.full(
            [number_of_frames[segment], number_of_home_players * 2], np.nan
        )
        for segment in segments
    }
    xydata["Away"] = {
        segment: np.full(
            [number_of_frames[segment], number_of_away_players * 2], np.nan
        )
        for segment in segments
    }
    xydata["Ball"] = {
        segment: np.full([number_of_frames[segment], 2], np.nan) for segment in segments
    }

    # read txt file from disk
    with open(filepath_txt, "r") as f:
        tracking_data_lines = f.readlines()

    # loop
    for package in tracking_data_lines:

        # read line to get gameclock, player positions and ball info
        (
            gameclock,
            segment,
            positions,
            ball,
        ) = _read_tracking_data_txt_single_line(package)

        # check if frame is in any segment
        if segment is None:
            # skip line if not
            continue
        else:
            # otherwise calculate relative frame (in respective segment)
            frame_rel = int((gameclock - periods[segment][0]) / 1000 * framerate_est)

        # insert (x,y)-data into np.array
        for team in ["Home", "Away"]:
            for jID in positions[team].keys():
                # map jersey number to array index and infer respective columns
                x_col = (links[team][jID] - 1) * 2
                y_col = (links[team][jID] - 1) * 2 + 1
                xydata[team][segment][frame_rel, x_col] = positions[team][jID][0]
                xydata[team][segment][frame_rel, y_col] = positions[team][jID][1]

        # get ball data
        xydata["Ball"][segment][frame_rel] = ball.get("position", np.nan)

    # create XY objects
    home_ht1 = XY(xy=xydata["Home"][1], framerate=framerate_est)
    home_ht2 = XY(xy=xydata["Home"][2], framerate=framerate_est)
    away_ht1 = XY(xy=xydata["Away"][1], framerate=framerate_est)
    away_ht2 = XY(xy=xydata["Away"][2], framerate=framerate_est)
    ball_ht1 = XY(xy=xydata["Ball"][1], framerate=framerate_est)
    ball_ht2 = XY(xy=xydata["Ball"][2], framerate=framerate_est)

    data_objects = (home_ht1, home_ht2, away_ht1, away_ht2, ball_ht1, ball_ht2)

    return data_objects


def read_event_data_from_url(
    url: str,
) -> Tuple[Events, Events, Events, Events, Pitch]:
    """Reads a URL containing a StatsPerform events csv file and extracts the stored
    event data and pitch information.

    The event data from the URL is downloaded into a temporary file stored in the
    repository's internal root ``.data``-folder and removed afterwards.

    Parameters
    ----------
    url: str
        URL to the xml file containing the event data.

    Returns
    -------
    data_objects: Tuple[Events, Events, Events, Events, Pitch]
        Events-objects for both teams and both halves and pitch information. The order
        is (home_ht1, home_ht2, away_ht1, away_ht2, pitch).
    """
    data_dir = os.path.join(DATA_DIR, "statsperform")
    if not os.path.isdir(data_dir):
        os.makedirs(data_dir, exist_ok=True)
    temp_file = os.path.join(data_dir, "events_temp.xml")
    with open(temp_file, "wb") as binary_file:
        binary_file.write(download_from_url(url))
    data_objects = read_event_data_xml(filepath_xml=os.path.join(data_dir, temp_file))
    os.remove(os.path.join(data_dir, temp_file))
    return data_objects


def read_tracking_data_from_url(
    url: str,
    links: Dict[str, Dict[int, int]] = None,
) -> Tuple[XY, XY, XY, XY, XY, XY]:
    """Reads a URL from the StatsPerform API (StatsEdgeViewer) containing a tracking
    data txt file and extracts position data.

    The tracking data from the URL is downloaded into a temporary file stored in the
    repository's internal root ``.data``-folder and removed afterwards.

    Parameters
    ----------
    url: str or pathlib.Path
        URL to the txt file containing the tracking data.
    links: Dict[str, Dict[int, int]], optional
        A link dictionary of the form ``links[team][jID] = xID``. Player's are
        identified in StatsPerform files via jID, and this dictionary is used to map
        them to a specific xID in the respective XY objects. Should be supplied if that
        order matters. If None is given (default), the links are automatically extracted
        from the csv file at the cost of a second pass through the entire file.

    Returns
    -------
    data_objects: Tuple[XY, XY, XY, XY, XY, XY]
        XY- and Pitch-objects for both teams and both halves. The order is
        (home_ht1, home_ht2, away_ht1, away_ht2, ball_ht1, ball_ht2, pitch)
    """
    data_dir = os.path.join(DATA_DIR, "statsperform")
    if not os.path.isdir(data_dir):
        os.makedirs(data_dir, exist_ok=True)
    temp_file = os.path.join(data_dir, "tracking_temp.txt")
    with open(temp_file, "wb") as binary_file:
        binary_file.write(download_from_url(url))
    data_objects = read_tracking_data_txt(
        filepath_txt=os.path.join(data_dir, temp_file),
        links=links,
    )
    os.remove(os.path.join(data_dir, temp_file))
    return data_objects



import numpy as np

from floodlight import XY
from floodlight.core.property import PlayerProperty
from floodlight.models.base import BaseModel, requires_fit


class DistanceModel(BaseModel):
    """Computations for Euclidean distances of all players.

    Upon calling the :func:`~DistanceModel.fit`-method, this model calculates the
    frame-wise Euclidean distance for each player. The following calculations can
    subsequently be queried by calling the corresponding methods:

        - Frame-wise Distance Covered --> :func:`~DistanceModel.distance_covered`
        - Cumulative Distance Covered --> :func:`~DistanceModel.cumulative_distance_\
covered`

    Notes
    -----
    For input data in metrical units, the output equals the input unit.
    Differences between frames can be calculated with two different methods:

        *Central difference method* (recommended) allows for differenciation without
        temporal shift:

                .. math::

                    y^{\\prime}(t_{0}) = \\frac{y_{1}-y_{-1}}{t_{1} - t_{-1}}

        The first and last frame are padded with linear extrapolation.\n
        *Backward difference method* calculates the difference between each consecutive
        frame:

                .. math::

                    y^{\\prime}(t_{0}) = \\frac{y_{0}-y_{-1}}{t_{0} - t_{-1}}

        The first frame is padded prepending a '0' at the beginning of the array along
        axis=1.

    Examples
    --------
    >>> import numpy as np
    >>> from floodlight import XY
    >>> from floodlight.models.kinematics import DistanceModel

    >>> xy = XY(np.array(((0, 0), (0, 1), (1, 1), (2, 2))))

    >>> dm = DistanceModel()
    >>> dm.fit(xy)

    >>> dm.distance_covered()
    PlayerProperty(property=array([[1.        ],
       [0.70710678],
       [1.11803399],
       [1.41421356]]), name='distance_covered'0)
    >>> dm.cumulative_distance_covered()
    PlayerProperty(property=array([[1.        ],
       [0.70710678],
       [1.11803399],
       [1.41421356]]), name='distance_covered')
    """

    def __init__(self):
        super().__init__()
        self._distance_euclidean_ = None

    def fit(self, xy: XY, difference: str = "central", axis: str = None):
        """Fits a model calculating Euclidean distances of each player to an XY object.

        Parameters
        ----------
        xy: XY
            Floodlight XY Data object.
        difference: {'central', 'backward'}, optional
            The method of differentiation. 'central' will differentiate using the
            central difference method, 'backward' will differentiate using the backward
            difference method as described in the Notes.
        axis: {None, 'x', 'y'}, optional
            Optional argument that restricts distance calculation to either the x-
            or y-dimension of the data. If set to None (default), distances are
            calculated in both dimensions.
        """

        if axis is None:
            if difference == "central":
                differences_xy = np.gradient(xy.xy, axis=0)
            elif difference == "backward":
                differences_xy = np.diff(xy.xy, axis=0, prepend=0)
            else:
                raise ValueError(
                    f"Expected axis to be one of (None, 'x', 'y'), got {axis}."
                )
            distance_euclidean = np.hypot(
                differences_xy[:, ::2],
                differences_xy[:, 1::2],
            )
        elif axis == "x":
            distance_euclidean = np.gradient(xy.x, axis=0)
        elif axis == "y":
            distance_euclidean = np.gradient(xy.y, axis=0)
        else:
            raise ValueError(
                f"Expected axis to be one of (None, 'x', 'y'), got {axis}."
            )

        self._distance_euclidean_ = PlayerProperty(
            property=distance_euclidean, name="distance_covered", framerate=xy.framerate
        )

    @requires_fit
    def distance_covered(self) -> PlayerProperty:
        """Returns the frame-wise distance covered as computed by the fit method.

        Returns
        -------
        distance_euclidean: PlayerProperty
            A PlayerProperty object of shape (T, N), where T is the total number of
            frames and N is the number of players. The columns contain the frame-wise
            Euclidean distance covered.
        """
        return self._distance_euclidean_

    @requires_fit
    def cumulative_distance_covered(self) -> PlayerProperty:
        """Returns the cumulative distance covered.

        Returns
        -------
        cumulative_distance_euclidean: PlayerProperty
            A PlayerProperty object of shape (T, N), where T is the total number of
            frames and N is the number of players. The columns contain the cumulative
            Euclidean distance covered calculated by numpy.nancumsum() over axis=0.
        """
        dist = self._distance_euclidean_.property
        cum_dist = np.nancumsum(dist, axis=0)
        cumulative_distance = PlayerProperty(
            property=cum_dist,
            name="cumulative_distance_covered",
            framerate=self._distance_euclidean_.framerate,
        )
        return cumulative_distance


class VelocityModel(BaseModel):
    """Computations for velocities of all players.

    Upon calling the :func:`~VelocityModel.fit`-method, this model calculates the
    frame-wise velocity for each player. The calculation can subsequently be queried by
    calling the corresponding method:

        - Frame-wise velocity --> :func:`~VelocityModel.velocity`

    Notes
    -----
    For input data in metrical units, the output equals the input unit.
    Differences between frames can be calculated with two different methods:

        *Central difference method* (recommended) allows for differenciation without
        temporal shift:

                .. math::

                    y^{\\prime}(t_{0}) = \\frac{y_{1}-y_{-1}}{t_{1} - t_{-1}}

        The first and last frame are padded with linear extrapolation.\n
        *Backward difference method* calculates the difference between each consecutive
        frame:

                .. math::

                    y^{\\prime}(t_{0}) = \\frac{y_{0}-y_{-1}}{t_{0} - t_{-1}}

        The first frame is padded prepending a '0' at the beginning of the array along
        axis=1.

    Examples
    --------
    >>> import numpy as np
    >>> from floodlight import XY
    >>> from floodlight.models.kinematics import VelocityModel

    >>> xy = XY(np.array(((0, 0), (0, 1), (1, 1), (2, 2))), framerate=20)

    >>> vm = VelocityModel()
    >>> vm.fit(xy)

    >>> vm.velocity()
    PlayerProperty(property=array([[20.        ],
       [14.14213562],
       [22.36067977],
       [28.28427125]]), name='velocity', framerate=20)
    """

    def __init__(self):
        super().__init__()
        self._velocity_ = None

    def fit(
        self,
        xy: XY,
        difference: str = "central",
        axis: str = None,
    ):
        """Fits a model calculating velocities of each player to an XY object.

        Parameters
        ----------
        xy: XY
            Floodlight XY Data object.
        difference: {'central', 'backward'}, optional
            The method of differentiation. 'central' will differentiate using the
            central difference method, 'backward' will differentiate using the backward
            difference method as described in the Notes.
        axis: {None, 'x', 'y'}, optional
            Optional argument that restricts distance calculation to either the x-
            or y-dimension of the data. If set to None (default), distances are
            calculated in both dimensions.
        """

        distance_model = DistanceModel()
        distance_model.fit(xy, difference=difference, axis=axis)
        distance_euclidean = distance_model.distance_covered()

        velocity = distance_euclidean.property * distance_euclidean.framerate

        self._velocity_ = PlayerProperty(
            property=velocity, name="velocity", framerate=xy.framerate
        )

    @requires_fit
    def velocity(self) -> PlayerProperty:
        """Returns the frame-wise velocity as computed by the fit method.

        Returns
        -------
        velocity: PlayerProperty
            A PlayerProperty object of shape (T, N), where T is the total number of
            frames and N is the number of players. The columns contain the frame-wise
            velocity.
        """
        return self._velocity_


class AccelerationModel(BaseModel):
    """Computations for velocities of all players.

    Upon calling the :func:`~AccelerationModel.fit`-method, this model calculates the
    frame-wise acceleration for each player. The calculation can subsequently be queried
    by calling the corresponding method:

        - Frame-wise acceleration --> :func:`~AccelerationModel.acceleration`

    Notes
    -----
    For input data in metrical units, the output equals the input unit.
    Differences between frames can be calculated with two different methods:

        *Central difference method* (recommended) allows for differenciation without
        temporal shift:

                .. math::

                    y^{\\prime}(t_{0}) = \\frac{y_{1}-y_{-1}}{t_{1} - t_{-1}}

        The first and last frame are padded with linear extrapolation.\n
        *Backward difference method* calculates the difference between each consecutive
        frame:

                .. math::

                    y^{\\prime}(t_{0}) = \\frac{y_{0}-y_{-1}}{t_{0} - t_{-1}}

        The first frame is padded prepending a '0' at the beginning of the array along
        axis=1.

    Examples
    --------
    >>> import numpy as np
    >>> from floodlight import XY
    >>> from floodlight.models.kinematics import AccelerationModel

    >>> xy = XY(np.array(((0, 0), (0, 1), (1, 1), (2, 2))), framerate=20)

    >>> am = AccelerationModel()
    >>> am.fit(xy)

    >>> am.acceleration()
    PlayerProperty(property=array([[-117.15728753],
       [  23.60679775],
       [ 141.42135624],
       [ 118.47182945]]), name='acceleration', framerate=20)
    """

    def __init__(self):
        super().__init__()
        self._acceleration_ = None

    def fit(
        self,
        xy: XY,
        difference: str = "central",
        axis: str = None,
    ):
        """Fits a model calculating accelerations of each player to an XY object.

        Parameters
        ----------
        xy: XY
            Floodlight XY Data object.
        difference: {'central', 'backward'}, optional
            The method of differentiation. 'central' will differentiate using the
            central difference method, 'backward' will differentiate using the backward
            difference method as described in the Notes.
        axis: {None, 'x', 'y'}, optional
            Optional argument that restricts distance calculation to either the x-
            or y-dimension of the data. If set to None (default), distances are
            calculated in both dimensions.
        """

        velocity_model = VelocityModel()
        velocity_model.fit(xy, difference=difference, axis=axis)
        velocity = velocity_model.velocity()

        if difference == "central":
            acceleration = np.gradient(velocity.property, axis=0) * velocity.framerate
        else:
            acceleration = (
                np.diff(velocity.property, axis=0, append=0) * velocity.framerate
            )

        self._acceleration_ = PlayerProperty(
            property=acceleration,
            name="acceleration",
            framerate=xy.framerate,
        )

    @requires_fit
    def acceleration(self) -> PlayerProperty:
        """Returns the frame-wise acceleration as computed by the fit method.

        Returns
        -------
        acceleration: PlayerProperty
            A PlayerProperty object of shape (T, N), where T is the total number of
            frames and N is the number of players. The columns contain the frame-wise
            acceleration.
        """
        return self._acceleration_



import re
from pathlib import Path
from typing import Tuple, Union

import pytz
import iso8601
import pandas as pd
from lxml import etree

from floodlight.core.events import Events
from floodlight.core.pitch import Pitch
from floodlight.io.utils import get_and_convert


def get_opta_feedtype(filepath: Union[str, Path]) -> Union[str, None]:
    """Tries to extract the feed type from Opta's XML feed.

    This function assumes that the file follows Opta's format of producing feeds.
    Thus it should have a "PRODUCTION HEADER" comment at the top of the file so that on
    line 6 it reads something like ``production module:  Opta::Feed::XML::Soccer::F24``.

    Parameters
    ----------
    filepath : Union[str, Path]
        Full path to Opta XML file.

    Returns
    -------
    feedtype: str or None
        Returns the type of the feed  as a string in case it  finds it, e.g. 'F24',
        and `None` otherwise.
    """
    with open(str(filepath), "r") as f:
        # iterate through first lines instead of loading entire file to RAM
        for i, line in enumerate(f):
            # search for production module at line 6
            if i == 6:
                production_tags = line.strip().split(":")
                if production_tags[0] == "production module":
                    feedtype = production_tags[-1]
                else:
                    feedtype = None
                break

    return feedtype


def read_f24(
    filepath: Union[str, Path]
) -> Tuple[Events, Events, Events, Events, Pitch]:
    """Parse Opta's f24 feed (containing match events) and extract event data and pitch
    information.

    This function provides a high-level access to the particular f24 feed and will
    return event objects for both teams. The number of segments is inferred from the
    data, yet data for each segment is stored in a separate object.

    Parameters
    ----------
    filepath: str or pathlib.Path
        Full path to the XML feed.

    Returns
    -------
    data_objects: Tuple[Events, Events, Events, Events, Pitch]
        Events- and Pitch-objects for both teams and both halves. The order is
        (home_ht1, home_ht2, away_ht1, away_ht2, pitch)

    Notes
    -----
    Opta's format of handling event data information involves an elaborate use of so
    called qualifiers, which attach additional information to certain events. There
    also exist a number of mappings that define which qualifiers may be  attached to
    which kind of events. Parsing this information involves quite a bit of logic and is
    planned to be inclucded in further releases. As of now, qualifier information is
    parsed as a string in the `qualifier` column of the returned DataFrame and can be
    transformed to a dict of the form `{qualifier_id: value}`.
    """
    # check feed type
    if get_opta_feedtype(filepath) != "F24":
        raise ValueError(f"Not an Opta F24 feed: {filepath}")

    # load xml tree into memory
    tree = etree.parse(str(filepath))
    root = tree.getroot()

    # 1. parse match info
    matchinfo = root.xpath("Game")[0].attrib
    teams = ["Home", "Away"]
    tID_link = {
        int(matchinfo["home_team_id"]): "Home",
        int(matchinfo["away_team_id"]): "Away",
    }
    number_of_periods = len(list(filter(re.compile("period_._start").match, matchinfo)))
    segments = [f"HT{period}" for period in range(1, number_of_periods + 1)]

    # 2. parse events
    # bins
    columns = [
        "eID",
        "gameclock",
        "pID",
        "outcome",
        "timestamp",
        "minute",
        "second",
        "at_x",
        "at_y",
        "qualifier",
    ]

    event_lists = {
        team: {segment: {col: [] for col in columns} for segment in segments}
        for team in teams
    }
    directions = {team: {} for team in teams}
    dir_link = {"Left to Right": "lr", "Right to Left": "rl"}
    segment_offsets = {1: 0, 2: 45, 3: 90, 4: 105}
    kickoffs = {}

    # read kickoff events for times and playing direction
    # (NOTE: kickoff times can also be directly found in matchinfo, although the
    # explicit kickoff-event timestamps appear to be more accurate)
    for event in root.xpath("Game/Event[@type_id='32']"):
        # get team and segment information
        period = get_and_convert(event.attrib, "period_id", int)
        segment = "HT" + str(period)
        tID = get_and_convert(event.attrib, "team_id", int)
        team = tID_link[tID]
        # read kickoff times
        kickoff_timestring = get_and_convert(event.attrib, "timestamp", str)
        kickoff_datetime = iso8601.parse_date(
            kickoff_timestring, default_timezone=pytz.utc
        )
        kickoffs[segment] = kickoff_datetime
        # read playing direction
        direction_qualifier = event.xpath("Q[@qualifier_id='127']")
        if len(direction_qualifier) > 0:
            value = get_and_convert(direction_qualifier[0], "value", str)
            direction = dir_link.get(value)
        else:
            direction = None
        directions[team][segment] = direction
        # cut event from tree to prevent double parsing
        # event.getparent().remove(event)

    # loop
    for event in root.xpath("Game/Event"):
        # get team and segment information
        period = get_and_convert(event.attrib, "period_id", int)
        segment = "HT" + str(period)
        tID = get_and_convert(event.attrib, "team_id", int)
        team = tID_link[tID]
        # skip match-unrelated events
        if period not in range(1, 6):
            continue

        # identifier and outcome:
        eID = get_and_convert(event.attrib, "type_id", int)
        # skip unwanted events
        if eID in [30]:
            continue
        pID = get_and_convert(event.attrib, "player_id", int)
        outcome = get_and_convert(event.attrib, "outcome", int)
        event_lists[team][segment]["eID"].append(eID)
        event_lists[team][segment]["pID"].append(pID)
        event_lists[team][segment]["outcome"].append(outcome)

        # absolute and relative time
        event_timestring = get_and_convert(event.attrib, "timestamp", str)
        minute = get_and_convert(event.attrib, "min", int)
        # transform minute to be relative to current segment
        minute -= segment_offsets[period]
        second = get_and_convert(event.attrib, "sec", int)
        timestamp = iso8601.parse_date(event_timestring, default_timezone=pytz.utc)
        delta = timestamp - kickoffs[segment]
        gameclock = delta.total_seconds()
        # re-adjust pre-kick-off events (e.g. substitutions) to 00:00
        gameclock = max(gameclock, 0.0)
        event_lists[team][segment]["timestamp"].append(timestamp)
        event_lists[team][segment]["minute"].append(minute)
        event_lists[team][segment]["second"].append(second)
        event_lists[team][segment]["gameclock"].append(gameclock)

        # location
        at_x = get_and_convert(event.attrib, "x", float)
        at_y = get_and_convert(event.attrib, "y", float)
        event_lists[team][segment]["at_x"].append(at_x)
        event_lists[team][segment]["at_y"].append(at_y)

        # qualifier
        qual_dict = {}
        for qualifier in event.iterchildren():
            qual_id = int(qualifier.attrib["qualifier_id"])
            qual_value = qualifier.attrib.get("value")
            qual_dict[qual_id] = qual_value
        event_lists[team][segment]["qualifier"].append(str(qual_dict))

    # assembly
    home_ht1 = Events(
        events=pd.DataFrame(data=event_lists["Home"]["HT1"]),
        direction=directions["Home"]["HT1"],
    )
    home_ht2 = Events(
        events=pd.DataFrame(data=event_lists["Home"]["HT2"]),
        direction=directions["Home"]["HT2"],
    )
    away_ht1 = Events(
        events=pd.DataFrame(data=event_lists["Away"]["HT1"]),
        direction=directions["Away"]["HT1"],
    )
    away_ht2 = Events(
        events=pd.DataFrame(data=event_lists["Away"]["HT2"]),
        direction=directions["Away"]["HT2"],
    )
    pitch = Pitch.from_template("opta", sport="football")

    data_objects = (home_ht1, home_ht2, away_ht1, away_ht2, pitch)

    return data_objects



import warnings

import numpy as np
from scipy.spatial.distance import cdist

from floodlight import XY
from floodlight.core.property import TeamProperty, PlayerProperty
from floodlight.models.base import BaseModel, requires_fit


class CentroidModel(BaseModel):
    """Computations based on the geometric center of all players, commonly referred to
    as a team's *centroid*.

    Upon calling the :func:`~CentroidModel.fit`-method, this model calculates a team's
    centroid. The following calculations can subsequently be queried by calling the
    corresponding methods:

        - Centroid [1]_ --> :func:`~CentroidModel.centroid`
        - Centroid Distance --> :func:`~CentroidModel.centroid_distance`
        - Stretch Index [2]_ --> :func:`~CentroidModel.stretch_index`

    Notes
    -----
    Team centroids are computed as the arithmetic mean of all player positions (based on
    *numpy*'s nanmean function). For a fixed point in time and :math:`N` players with
    corresponding positions :math:`x_1, \\dots, x_N \\in \\mathbb{R}^2`, the centroid is
    calculated as

        .. math::
            C = \\frac{1}{N} \\sum_i^N x_i.

    Examples
    --------
    >>> import numpy as np
    >>> from floodlight import XY
    >>> from floodlight.models.geometry import CentroidModel

    >>> xy = XY(np.array(((1, 1, 2, -2), (1.5, np.nan, np.nan, -0))))
    >>> cm = CentroidModel()
    >>> cm.fit(xy)
    >>> cm.centroid()
    XY(xy=array([[ 1.5, -0.5],
       [ 1.5,  0. ]]), framerate=None, direction=None)
    >>> cm.stretch_index(xy)
    TeamProperty(property=array([1.5811388, nan]), name='stretch_index', framerate=None)
    >>> cm.stretch_index(xy, axis='x')
    TeamProperty(property=array([0.5, 0.]), name='stretch_index', framerate=None)

    Attributes
    ----------
    _centroid_: XY
        Calculated centroids with shape (T, 2), where T is the total number of frames.

    References
    ----------
        .. [1] `Sampaio, J., & Maçãs, V. (2012). Measuring tactical behaviour in
            football. International Journal of Sports Medicine, 33(05), 395-401.
            <https://www.thieme-connect.de/products/ejournals/abstract/10.1055/s-0031-13
            01320>`_
        .. [2] `Bourbousson, J., Sève, C., & McGarry, T. (2010). Space–time coordination
            dynamics in basketball: Part 2. The interaction between the two teams.
            Journal of Sports Sciences, 28(3), 349-358.
            <https://www.tandfonline.com/doi/full/10.1080/02640410903503640>`_
    """

    def __init__(self):
        super().__init__()
        # model parameter
        self._centroid_ = None

    def fit(self, xy: XY, exclude_xIDs: list = None):
        """Fit the model to the given data and calculate team centroids.

        Parameters
        ----------
        xy: XY
            Player spatiotemporal data for which the centroid is calculated.
        exclude_xIDs: list, optional
            A list of xIDs to be excluded from computation. This can be useful if one
            would like, for example, to exclude goalkeepers from analysis.
        """
        if not exclude_xIDs:
            exclude_xIDs = []
        # boolean for column inclusion, initialize to True for all columns
        include = np.full((xy.N * 2), True)

        # exclude columns according to exclude_xIDs
        for xID in exclude_xIDs:
            if xID not in range(0, xy.N):
                raise ValueError(
                    f"Expected entries of exclude_xIDs to be in range 0 to {xy.N}, "
                    f"got {xID}."
                )
            exclude_start = xID * 2
            exclude_end = exclude_start + 2
            include[exclude_start:exclude_end] = False

        with warnings.catch_warnings():
            # supress warnings caused by empty slices
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            # calculate centroid
            centroids = np.nanmean(xy.xy[:, include].reshape((len(xy), -1, 2)), axis=1)

        # wrap as XY object
        self._centroid_ = XY(
            xy=centroids, framerate=xy.framerate, direction=xy.direction
        )

    @requires_fit
    def centroid(self) -> XY:
        """Returns the team centroid positions as computed by the fit method.

        Returns
        -------
        centroid: XY
            An XY object of shape (T, 2), where T is the total number of frames. The two
            columns contain the centroids' x- and y-coordinates, respectively.
        """
        return self._centroid_

    @requires_fit
    def centroid_distance(self, xy: XY, axis: str = None) -> PlayerProperty:
        """Calculates the Euclidean distance of each player to the fitted centroids.

        Parameters
        ----------
        xy: XY
            Player spatiotemporal data for which the distances to the fitted centroids
            are calculated.
        axis: {None, 'x', 'y'}, optional
            Optional argument that restricts distance calculation to either the x- or
            y-dimension of the data. If set to None (default), distances are calculated
            in both dimensions.

        Returns
        -------
        centroid_distance: PlayerProperty
            A PlayerProperty object of shape (T, N), where T is the total number of
            frames. Each column contains the distances to the team centroid of the
            player with corresponding xID.
        """
        # check matching lengths
        T = len(self._centroid_)
        if len(xy) != T:
            raise ValueError(
                f"Length of xy ({len(xy)}) does not match length of fitted centroids "
                f"({T})."
            )

        # calculate distances on specified axis
        distances = np.full((T, xy.N), np.nan)
        if axis is None:
            for t in range(T):
                distances[t] = cdist(
                    self._centroid_[t].reshape(-1, 2), xy[t].reshape(-1, 2)
                )
        elif axis == "x":
            for t in range(T):
                distances[t] = cdist(
                    self._centroid_.x[t].reshape(-1, 1), xy.x[t].reshape(-1, 1)
                )
        elif axis == "y":
            for t in range(T):
                distances[t] = cdist(
                    self._centroid_.y[t].reshape(-1, 1), xy.y[t].reshape(-1, 1)
                )
        else:
            raise ValueError(
                f"Expected axis to be one of (None, 'x', 'y'), got {axis}."
            )

        # wrap as PlayerProperty
        centroid_distance = PlayerProperty(
            property=distances,
            name="centroid_distance",
            framerate=xy.framerate,
        )

        return centroid_distance

    @requires_fit
    def stretch_index(self, xy: XY, axis: str = None) -> TeamProperty:
        """Calculates the *Stretch Index*, i.e., the mean distance of all players to the
        team centroid.

        Parameters
        ----------
        xy: XY
            Player spatiotemporal data for which the stretch index is calculated.
        axis: {None, 'x', 'y'}, optional
            Optional argument that restricts stretch index calculation to either the x-
            or y-dimension of the data. If set to None (default), the stretch index is
            calculated in both dimensions.

        Returns
        -------
        stretch_index: TeamProperty
            A TeamProperty object of shape (T, 1), where T is the total number of
            frames. Each entry contains the stretch index of that particular frame.
        """
        # get player distances from centroid
        centroid_distances = self.centroid_distance(xy=xy, axis=axis)

        with warnings.catch_warnings():
            # supress warnings caused by empty slices
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            # calculate stretch index
            stretch_index = np.nanmean(centroid_distances.property, axis=1)

        # wrap as TeamProperty object
        stretch_index = TeamProperty(
            property=stretch_index, name="stretch_index", framerate=xy.framerate
        )

        return stretch_index



import warnings
from dataclasses import dataclass
from typing import Tuple

import matplotlib
import matplotlib.pyplot as plt

from floodlight.vis.pitches import plot_handball_pitch, plot_football_pitch
from floodlight.utils.types import Numeric


@dataclass
class Pitch:
    """
    Pitch and coordinate system specifications. Core class of floodlight.

    Attributes
    ----------
    xlim: Tuple[Numeric, Numeric]
        Limits of pitch boundaries in longitudinal direction. This tuple has the form
        (x_min, x_max) and delimits the length of the pitch (not of any actual data)
        within the coordinate system.
    ylim: Tuple[Numeric, Numeric]
        Limits of pitch boundaries in lateral direction. This tuple has the form
        (y_min, y_max) and delimits the width of the pitch (not of any actual data)
        within the coordinate system.
    unit: str
        The unit in which data is measured along axes. Possible types are
        {'m', 'cm', 'percent'}, whereas 'percent' should be used to denote standardized
        pitches where data is scaled along the axes independent of the actual pitch
        size. In this case, you should specify the `length` and `width` attributes.
    boundaries: str
        One of {'fixed', 'flexible'}. Here, 'fixed' denotes coordinate systems that
        limit axes to the respective xlim and ylim. On the contrary, 'flexible'
        coordinate systems do not explicitly specify a limit. Instead, the limit is
        implicitly set by the actual pitch length and width.
    length: Numeric, optional
        Actual pitch length in *m*.
    width: Numeric, optional
        Actual pitch width in *m*.
    sport: str, optional
        Sport for which the pitch is used. This is used to automatically generate lines
        and markings.
    """

    xlim: Tuple[Numeric, Numeric]
    ylim: Tuple[Numeric, Numeric]
    unit: str
    boundaries: str
    length: Numeric = None
    width: Numeric = None
    sport: str = None

    def __str__(self):
        return (
            f"Floodlight Pitch object with axes x = {self.xlim} / y = {self.ylim} "
            f"({self.boundaries}) in [{self.unit}]"
        )

    @classmethod
    def from_template(cls, template_name: str, **kwargs):
        """
        Creates a Pitch object representing common data provider formats.

        Parameters
        ----------
        template_name: str
            The name of the template the pitch should follow. Currently supported are
            {'dfl', 'eigd', 'opta', 'statsbomb', 'statsperform', 'statsperform_open',
            'tracab'}.
        kwargs:
            You may pass optional arguments (`length`, `width`, `sport`) used for class
            instantiation. For some data providers, additional kwargs are needed to
            represent their format correctly. For example, pass the `length` and `width`
            argument to create a Pitch object in the 'tracab' format.

        Returns
        -------
        pitch: Pitch
            A class instance of the given provider format.
        """
        if template_name == "dfl":
            if "length" not in kwargs or "width" not in kwargs:
                raise TypeError(
                    "For an exact DFL (German Football League) "
                    "Pitch object, `length` and `width` of the pitch need"
                    "to be passed as keyworded arguments"
                )
            x_half = round(kwargs["length"] / 2, 3)
            y_half = round(kwargs["width"] / 2, 3)
            return cls(
                xlim=(-x_half, x_half),
                ylim=(-y_half, y_half),
                unit="m",
                boundaries="flexible",
                length=kwargs.get("length"),
                width=kwargs.get("width"),
                sport=kwargs.get("sport"),
            )
        elif template_name == "opta":
            return cls(
                xlim=(0.0, 100.0),
                ylim=(0.0, 100.0),
                unit="percent",
                boundaries="fixed",
                length=kwargs.get("length"),
                width=kwargs.get("width"),
                sport=kwargs.get("sport"),
            )
        elif template_name == "statsperform_open":
            if "length" not in kwargs or "width" not in kwargs:
                raise TypeError(
                    "For an exact StatsPerform Pitch object, "
                    "`length` and `width` of the pitch need "
                    "to be passed as keyworded arguments"
                )
            x_half = round(kwargs["length"] / 2, 3)
            y_half = round(kwargs["width"] / 2, 3)
            return cls(
                xlim=(-x_half, x_half),
                ylim=(-y_half, y_half),
                unit="m",
                boundaries="flexible",
                length=kwargs.get("length"),
                width=kwargs.get("width"),
                sport=kwargs.get("sport"),
            )
        elif template_name == "statsperform":
            if "length" not in kwargs or "width" not in kwargs:
                raise TypeError(
                    "For an exact StatsPerform Pitch object, "
                    "`length` and `width` of the pitch need "
                    "to be passed as keyworded arguments"
                )
            x_half = round(kwargs["length"] / 2, 3)
            y_half = round(kwargs["width"] / 2, 3)
            return cls(
                xlim=(-x_half, x_half),
                ylim=(-y_half, y_half),
                unit="cm",
                boundaries="flexible",
                length=kwargs.get("length"),
                width=kwargs.get("width"),
            )
        elif template_name == "tracab":
            if "length" not in kwargs or "width" not in kwargs:
                raise TypeError(
                    "For an exact TRACAB (ChyronHego, international format) "
                    "Pitch object, `length` and `width` of the pitch need "
                    "to be passed as keyworded arguments"
                )
            x_half = round(kwargs["length"] / 2, 3)
            y_half = round(kwargs["width"] / 2, 3)
            return cls(
                xlim=(-x_half, x_half),
                ylim=(-y_half, y_half),
                unit="cm",
                boundaries="flexible",
                length=kwargs.get("length"),
                width=kwargs.get("width"),
                sport=kwargs.get("sport"),
            )
        elif template_name == "eigd":
            return cls(
                xlim=(0, 40),
                ylim=(0, 20),
                unit="m",
                boundaries="fixed",
                length=40,
                width=20,
                sport="handball",
            )
        elif template_name == "statsbomb":
            return cls(
                xlim=(0.0, 120.0),
                ylim=(0.0, 80.0),
                unit="normed",
                boundaries="flexible",
                sport="football",
            )
        else:
            raise ValueError(f"Unsupported template name '{template_name}'")

    @property
    def is_metrical(self) -> bool:
        is_metrical = False
        if self.unit in ["m", "cm"]:
            is_metrical = True
        return is_metrical

    @property
    def center(self) -> tuple:
        center = (
            round((self.xlim[0] + self.xlim[1]) / 2, 3),
            round((self.ylim[0] + self.ylim[1]) / 2, 3),
        )
        return center

    def plot(
        self,
        color_scheme: str = "standard",
        show_axis_ticks: bool = False,
        ax: matplotlib.axes = None,
        **kwargs,
    ) -> matplotlib.axes:
        """Plots a pitch on a matplotlib.axes for a given sport.

        Parameters
        ----------
        color_scheme: str, optional
            Color scheme of the plot. One of {'standard', 'bw'}. Defaults to 'standard'.
        show_axis_ticks: bool, optional
            If set to True, the axis ticks are visible. Defaults to False.
        ax: matplotlib.axes, optional
            Axes from matplotlib library on which the playing field is plotted. If ax is
            None, a default-sized matplotlib.axes object is created.
        kwargs:
            Optional keyworded arguments {'linewidth', 'zorder', 'scalex', 'scaley'}
            which can be used for the plot functions from matplotlib. The kwargs are
            only passed to all the plot functions of matplotlib.

        Returns
        -------
        axes: matplotlib.axes
            Axes from matplotlib library on which the specified pitch is plotted.

        Notes
        -----
        The kwargs are only passed to the plot functions of matplotlib. To customize the
        plots have a look at `matplotlib
        <https://matplotlib.org/3.5.0/api/_as_gen/matplotlib.axes.Axes.plot.html>`_.
        For example in order to modify the linewidth pass a float to the keyworded
        argument 'linewidth'. The same principle applies to other kwargs like 'zorder',
        'scalex' and 'scaley'.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> from floodlight.core.pitch import Pitch
        >>> # create Pitch object
        >>> football_pitch = Pitch(xlim=(0, 105), ylim=(0,68), unit="m",
        >>>                  sport="football")

        >>> # plot football pitch
        >>> football_pitch.plot()
        >>> plt.show()

        .. image:: ../../_img/football_pitch_example.png

        >>> # create Pitch object
        >>> handball_pitch = Pitch(xlim=(0,40), ylim=(0,20), unit="m", sport="handball")

        >>> # plot handball pitch
        >>> handball_pitch.plot()
        >>> plt.show()

        .. image:: ../../_img/handball_pitch_example.png

        """
        # list of existing color_schemes and sports
        color_schemes = ["bw", "standard"]
        sports = ["football", "handball"]
        sport = self.sport

        # check if valide sport was chosen
        if sport not in sports or sport is None:
            raise ValueError(
                f"Expected self.sport to be from {sports}, got {self.sport}"
            )

        # check if a valide color scheme was chosen
        if color_scheme not in color_schemes:
            raise ValueError(
                f"Expected color_scheme to be from {color_schemes}, got {color_scheme}"
            )

        # check wether an axes to plot is given or if a new axes element has to be
        # created
        ax = ax or plt.subplots()[1]

        # set ratio between x and y values of the plot to ensure that the ratio between
        # length and width is correct regardless of the figsize.
        default_length = 105
        default_width = 68

        if self.unit != "percent":
            ax.set_aspect(1)
        # set ratio if unit is percent and sport is football
        elif self.unit == "percent" and sport == "football":
            if self.length and self.width:
                ax.set_aspect(self.width / self.length)
            # set ratio to standard pitch size of 68/105
            else:  # standard ratio of length and width
                ax.set_aspect(default_width / default_length)
                warnings.warn(
                    "Since self.unit == 'percent' but self.length and self.width are "
                    f"None the pitch is set to default values length: {default_length} "
                    f"and width: {default_width}"
                )
        # set ratio if unit is percent and sport is handball
        elif self.unit == "percent" and sport == "handball":
            ax.set_aspect(0.5)

        # create matplotlib.axes with handball pitch
        if sport == "handball":
            return plot_handball_pitch(
                self.xlim,
                self.ylim,
                self.unit,
                color_scheme=color_scheme,
                show_axis_ticks=show_axis_ticks,
                ax=ax,
                **kwargs,
            )

        # create matplotlib.axes with football pitch
        if sport == "football":
            return plot_football_pitch(
                self.xlim,
                self.ylim,
                self.length,
                self.width,
                self.unit,
                color_scheme=color_scheme,
                show_axis_ticks=show_axis_ticks,
                ax=ax,
                **kwargs,
            )
